@@ -18,6 +18,8 @@ typedef enum {
 
 void orc_test_init (void);
 OrcTestResult orc_test_gcc_compile (OrcProgram *p);
+OrcTestResult orc_test_gcc_compile_neon (OrcProgram *p);
+OrcTestResult orc_test_gcc_compile_c64x (OrcProgram *p);
 void orc_test_random_bits (void *data, int n_bytes);
 OrcTestResult orc_test_compare_output (OrcProgram *program);
 OrcTestResult orc_test_compare_output_full (OrcProgram *program, int flags);
@@ -27,6 +29,8 @@ OrcProgram *orc_test_get_program_for_opcode (OrcStaticOpcode *opcode);
 OrcProgram *orc_test_get_program_for_opcode_const (OrcStaticOpcode *opcode);
 OrcProgram *orc_test_get_program_for_opcode_param (OrcStaticOpcode *opcode);
 
+void orc_test_performance (OrcProgram *program, int flags);
+
 ORC_END_DECLS
 
 #endif
