@@ -1,113 +1,113 @@
 UNARY_SB(absb, "ORC_ABS(%s)")
 BINARY_SB(addb, "%s + %s")
 BINARY_SB(addssb, "ORC_CLAMP_SB(%s + %s)")
-BINARY_UB(addusb, "ORC_CLAMP_UB((uint8_t)%s + (uint8_t)%s)")
+BINARY_UB(addusb, "ORC_CLAMP_UB((orc_uint8)%s + (orc_uint8)%s)")
 BINARY_SB(andb, "%s & %s")
 BINARY_SB(andnb, "(~%s) & %s")
 BINARY_SB(avgsb, "(%s + %s + 1)>>1")
-BINARY_UB(avgub, "((uint8_t)%s + (uint8_t)%s + 1)>>1")
+BINARY_UB(avgub, "((orc_uint8)%s + (orc_uint8)%s + 1)>>1")
 BINARY_SB(cmpeqb, "(%s == %s) ? (~0) : 0")
 BINARY_SB(cmpgtsb, "(%s > %s) ? (~0) : 0")
 UNARY_SB(copyb, "%s")
 BINARY_SB(maxsb, "ORC_MAX(%s, %s)")
-BINARY_UB(maxub, "ORC_MAX((uint8_t)%s, (uint8_t)%s)")
+BINARY_UB(maxub, "ORC_MAX((orc_uint8)%s, (orc_uint8)%s)")
 BINARY_SB(minsb, "ORC_MIN(%s, %s)")
-BINARY_UB(minub, "ORC_MIN((uint8_t)%s, (uint8_t)%s)")
+BINARY_UB(minub, "ORC_MIN((orc_uint8)%s, (orc_uint8)%s)")
 BINARY_SB(mullb, "(%s * %s) & 0xff")
 BINARY_SB(mulhsb, "(%s * %s) >> 8")
-BINARY_UB(mulhub, "((uint32_t)(uint8_t)%s * (uint32_t)(uint8_t)%s) >> 8")
+BINARY_UB(mulhub, "((orc_uint32)(orc_uint8)%s * (orc_uint32)(orc_uint8)%s) >> 8")
 BINARY_SB(orb, "%s | %s")
 BINARY_SB(shlb, "%s << %s")
 BINARY_SB(shrsb, "%s >> %s")
-BINARY_UB(shrub, "((uint8_t)%s) >> %s")
+BINARY_UB(shrub, "((orc_uint8)%s) >> %s")
 UNARY_SB(signb, "ORC_CLAMP(%s,-1,1)")
 BINARY_SB(subb, "%s - %s")
 BINARY_SB(subssb, "ORC_CLAMP_SB(%s - %s)")
-BINARY_UB(subusb, "ORC_CLAMP_UB((uint8_t)%s - (uint8_t)%s)")
+BINARY_UB(subusb, "ORC_CLAMP_UB((orc_uint8)%s - (orc_uint8)%s)")
 BINARY_SB(xorb, "%s ^ %s")
 
 UNARY_SW(absw, "ORC_ABS(%s)")
 BINARY_SW(addw, "%s + %s")
 BINARY_SW(addssw, "ORC_CLAMP_SW(%s + %s)")
-BINARY_UW(addusw, "ORC_CLAMP_UW((uint16_t)%s + (uint16_t)%s)")
+BINARY_UW(addusw, "ORC_CLAMP_UW((orc_uint16)%s + (orc_uint16)%s)")
 BINARY_SW(andw, "%s & %s")
 BINARY_SW(andnw, "(~%s) & %s")
 BINARY_SW(avgsw, "(%s + %s + 1)>>1")
-BINARY_UW(avguw, "((uint16_t)%s + (uint16_t)%s + 1)>>1")
+BINARY_UW(avguw, "((orc_uint16)%s + (orc_uint16)%s + 1)>>1")
 BINARY_SW(cmpeqw, "(%s == %s) ? (~0) : 0")
 BINARY_SW(cmpgtsw, "(%s > %s) ? (~0) : 0")
 UNARY_SW(copyw, "%s")
 BINARY_SW(maxsw, "ORC_MAX(%s, %s)")
-BINARY_UW(maxuw, "ORC_MAX((uint16_t)%s, (uint16_t)%s)")
+BINARY_UW(maxuw, "ORC_MAX((orc_uint16)%s, (orc_uint16)%s)")
 BINARY_SW(minsw, "ORC_MIN(%s, %s)")
-BINARY_UW(minuw, "ORC_MIN((uint16_t)%s, (uint16_t)%s)")
+BINARY_UW(minuw, "ORC_MIN((orc_uint16)%s, (orc_uint16)%s)")
 BINARY_SW(mullw, "(%s * %s) & 0xffff")
 BINARY_SW(mulhsw, "(%s * %s) >> 16")
-BINARY_UW(mulhuw, "((uint32_t)((uint16_t)%s) * (uint32_t)((uint16_t)%s)) >> 16")
+BINARY_UW(mulhuw, "((orc_uint32)((orc_uint16)%s) * (orc_uint32)((orc_uint16)%s)) >> 16")
 BINARY_SW(orw, "%s | %s")
 BINARY_SW(shlw, "%s << %s")
 BINARY_SW(shrsw, "%s >> %s")
-BINARY_UW(shruw, "((uint16_t)%s) >> %s")
+BINARY_UW(shruw, "((orc_uint16)%s) >> %s")
 UNARY_SW(signw, "ORC_CLAMP(%s,-1,1)")
 BINARY_SW(subw, "%s - %s")
 BINARY_SW(subssw, "ORC_CLAMP_SW(%s - %s)")
-BINARY_UW(subusw, "ORC_CLAMP_UW((uint16_t)%s - (uint16_t)%s)")
+BINARY_UW(subusw, "ORC_CLAMP_UW((orc_uint16)%s - (orc_uint16)%s)")
 BINARY_SW(xorw, "%s ^ %s")
 
 UNARY_SL(absl, "ORC_ABS(%s)")
 BINARY_SL(addl, "%s + %s")
-BINARY_SL(addssl, "ORC_CLAMP_SL((int64_t)%s + (int64_t)%s)")
-BINARY_UL(addusl, "ORC_CLAMP_UL((int64_t)(uint32_t)%s + (int64_t)(uint32_t)%s)")
+BINARY_SL(addssl, "ORC_CLAMP_SL((orc_int64)%s + (orc_int64)%s)")
+BINARY_UL(addusl, "ORC_CLAMP_UL((orc_int64)(orc_uint32)%s + (orc_int64)(orc_uint32)%s)")
 BINARY_SL(andl, "%s & %s")
 BINARY_SL(andnl, "(~%s) & %s")
-BINARY_SL(avgsl, "((int64_t)%s + (int64_t)%s + 1)>>1")
-BINARY_UL(avgul, "((uint64_t)(uint32_t)%s + (uint64_t)(uint32_t)%s + 1)>>1")
+BINARY_SL(avgsl, "((orc_int64)%s + (orc_int64)%s + 1)>>1")
+BINARY_UL(avgul, "((orc_uint64)(orc_uint32)%s + (orc_uint64)(orc_uint32)%s + 1)>>1")
 BINARY_SL(cmpeql, "(%s == %s) ? (~0) : 0")
 BINARY_SL(cmpgtsl, "(%s > %s) ? (~0) : 0")
 UNARY_SL(copyl, "%s")
 BINARY_SL(maxsl, "ORC_MAX(%s, %s)")
-BINARY_UL(maxul, "ORC_MAX((uint32_t)%s, (uint32_t)%s)")
+BINARY_UL(maxul, "ORC_MAX((orc_uint32)%s, (orc_uint32)%s)")
 BINARY_SL(minsl, "ORC_MIN(%s, %s)")
-BINARY_UL(minul, "ORC_MIN((uint32_t)%s, (uint32_t)%s)")
+BINARY_UL(minul, "ORC_MIN((orc_uint32)%s, (orc_uint32)%s)")
 BINARY_SL(mulll, "(%s * %s) & 0xffffffff")
-BINARY_SL(mulhsl, "((int64_t)%s * (int64_t)%s) >> 32")
-BINARY_UL(mulhul, "((uint64_t)%s * (uint64_t)%s) >> 32")
+BINARY_SL(mulhsl, "((orc_int64)%s * (orc_int64)%s) >> 32")
+BINARY_UL(mulhul, "((orc_uint64)%s * (orc_uint64)%s) >> 32")
 BINARY_SL(orl, "%s | %s")
 BINARY_SL(shll, "%s << %s")
 BINARY_SL(shrsl, "%s >> %s")
-BINARY_UL(shrul, "((uint32_t)%s) >> %s")
+BINARY_UL(shrul, "((orc_uint32)%s) >> %s")
 UNARY_SL(signl, "ORC_CLAMP(%s,-1,1)")
 BINARY_SL(subl, "%s - %s")
-BINARY_SL(subssl, "ORC_CLAMP_SL((int64_t)%s - (int64_t)%s)")
-BINARY_UL(subusl, "ORC_CLAMP_UL((int64_t)(uint32_t)%s - (int64_t)(uint32_t)%s)")
+BINARY_SL(subssl, "ORC_CLAMP_SL((orc_int64)%s - (orc_int64)%s)")
+BINARY_UL(subusl, "ORC_CLAMP_UL((orc_int64)(orc_uint32)%s - (orc_int64)(orc_uint32)%s)")
 BINARY_SL(xorl, "%s ^ %s")
 
 UNARY_BW(convsbw, "%s")
-UNARY_BW(convubw, "(uint8_t)%s")
+UNARY_BW(convubw, "(orc_uint8)%s")
 UNARY_WL(convswl, "%s")
-UNARY_WL(convuwl, "(uint16_t)%s")
+UNARY_WL(convuwl, "(orc_uint16)%s")
 UNARY_WB(convwb, "%s")
 UNARY_WB(convssswb, "ORC_CLAMP_SB(%s)")
 UNARY_WB(convsuswb, "ORC_CLAMP_UB(%s)")
-UNARY_WB(convusswb, "ORC_CLAMP_SB((uint16_t)%s)")
-UNARY_WB(convuuswb, "ORC_CLAMP_UB((uint16_t)%s)")
+UNARY_WB(convusswb, "ORC_CLAMP_SB((orc_uint16)%s)")
+UNARY_WB(convuuswb, "ORC_CLAMP_UB((orc_uint16)%s)")
 UNARY_LW(convlw, "%s")
 UNARY_LW(convssslw, "ORC_CLAMP_SW(%s)")
 UNARY_LW(convsuslw, "ORC_CLAMP_UW(%s)")
-UNARY_LW(convusslw, "ORC_CLAMP_SW((uint32_t)%s)")
-UNARY_LW(convuuslw, "ORC_CLAMP_UW((uint32_t)%s)")
+UNARY_LW(convusslw, "ORC_CLAMP_SW((orc_uint32)%s)")
+UNARY_LW(convuuslw, "ORC_CLAMP_UW((orc_uint32)%s)")
 
 BINARY_BW(mulsbw, "%s * %s")
-BINARY_BW(mulubw, "(uint8_t)%s * (uint8_t)%s")
+BINARY_BW(mulubw, "(orc_uint8)%s * (orc_uint8)%s")
 BINARY_WL(mulswl, "%s * %s")
-BINARY_WL(muluwl, "(uint16_t)%s * (uint16_t)%s")
+BINARY_WL(muluwl, "(orc_uint16)%s * (orc_uint16)%s")
 
-BINARY_WL(mergewl, "((uint16_t)%s) | ((uint16_t)%s << 16)")
-BINARY_BW(mergebw, "((uint8_t)%s) | ((uint8_t)%s << 8)")
-UNARY_WB(select0wb, "(uint16_t)%s & 0xff")
-UNARY_WB(select1wb, "((uint16_t)%s >> 8)&0xff")
-UNARY_LW(select0lw, "(uint32_t)%s & 0xffff")
-UNARY_LW(select1lw, "((uint32_t)%s >> 16)&0xffff")
+BINARY_WL(mergewl, "((orc_uint16)%s) | ((orc_uint16)%s << 16)")
+BINARY_BW(mergebw, "((orc_uint8)%s) | ((orc_uint8)%s << 8)")
+UNARY_WB(select0wb, "(orc_uint16)%s & 0xff")
+UNARY_WB(select1wb, "((orc_uint16)%s >> 8)&0xff")
+UNARY_LW(select0lw, "(orc_uint32)%s & 0xffff")
+UNARY_LW(select1lw, "((orc_uint32)%s >> 16)&0xffff")
 UNARY_UW(swapw, "ORC_SWAP_W(%s)")
 UNARY_UL(swapl, "ORC_SWAP_L(%s)")
 
@@ -121,6 +121,6 @@ BINARY_F(minf, "ORC_MIN(%s,%s)")
 BINARY_FL(cmpeqf, "(%s == %s) ? (~0) : 0")
 BINARY_FL(cmpltf, "(%s < %s) ? (~0) : 0")
 BINARY_FL(cmplef, "(%s <= %s) ? (~0) : 0")
-UNARY_FL(convfl, "rintf(%s)")
+UNARY_FL(convfl, "(int)%s")
 UNARY_LF(convlf, "%s")
 
