@@ -35,6 +35,36 @@ void orc_compiler_rewrite_vars2 (OrcCompiler *compiler);
 int orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j);
 void orc_compiler_check_sizes (OrcCompiler *compiler);
 
+static char **_orc_compiler_flag_list;
+int _orc_compiler_flag_backup;
+int _orc_compiler_flag_debug;
+
+void
+_orc_compiler_init (void)
+{
+  const char *envvar;
+
+  envvar = getenv ("ORC_CODE");
+  if (envvar != NULL) {
+    _orc_compiler_flag_list = strsplit (envvar, ',');
+  }
+
+  _orc_compiler_flag_backup = orc_compiler_flag_check ("backup");
+  _orc_compiler_flag_debug = orc_compiler_flag_check ("debug");
+}
+
+int
+orc_compiler_flag_check (const char *flag)
+{
+  int i;
+
+  if (_orc_compiler_flag_list == NULL) return FALSE;
+
+  for (i=0;_orc_compiler_flag_list[i];i++){
+    if (strcmp (_orc_compiler_flag_list[i], flag) == 0) return TRUE;
+  }
+  return FALSE;
+}
 
 int
 orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
@@ -77,7 +107,8 @@ orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
   }
 
   /* FIXME on !x86, this is an error */
-  ORC_COMPILER_ERROR (compiler, "register overflow");
+  ORC_COMPILER_ERROR (compiler, "register overflow for %s reg",
+      data_reg ? "vector" : "gp");
   compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
 
   return 0;
@@ -122,7 +153,11 @@ orc_program_compile_for_target (OrcProgram *program, OrcTarget *target)
 {
   unsigned int flags;
 
-  flags = target->get_default_flags ();
+  if (target) {
+    flags = target->get_default_flags ();
+  } else {
+    flags = 0;
+  }
 
   return orc_program_compile_full (program, target, flags);
 }
@@ -158,6 +193,18 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
   compiler->target = target;
   compiler->target_flags = flags;
 
+  if (program->backup_func && _orc_compiler_flag_backup) {
+    ORC_COMPILER_ERROR(compiler, "Compilation disabled");
+    compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+    goto error;
+  }
+
+  if (target == NULL) {
+    ORC_COMPILER_ERROR(compiler, "No target given");
+    compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+    goto error;
+  }
+
   {
     ORC_LOG("variables");
     for(i=0;i<ORC_N_VARIABLES;i++){
@@ -299,7 +346,8 @@ orc_compiler_assign_rules (OrcCompiler *compiler)
         compiler->target_flags);
 
     if (insn->rule == NULL || insn->rule->emit == NULL) {
-      ORC_COMPILER_ERROR(compiler, "No rule for: %s", insn->opcode->name);
+      ORC_COMPILER_ERROR(compiler, "No rule for: %s on target %s",
+          insn->opcode->name, compiler->target->name);
       compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
       return;
     }
@@ -449,10 +497,17 @@ orc_compiler_global_reg_alloc (OrcCompiler *compiler)
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
         break;
     }
+
+    if (compiler->error) break;
   }
 
-  if (compiler->alloc_loop_counter) {
+  if (compiler->alloc_loop_counter && !compiler->error) {
     compiler->loop_counter = orc_compiler_allocate_register (compiler, FALSE);
+    /* FIXME massive hack */
+    if (compiler->loop_counter == 0) {
+      compiler->error = FALSE;
+      compiler->result = ORC_COMPILE_RESULT_OK;
+    }
   }
 }
 
