@@ -41,6 +41,9 @@
 #include <signal.h>
 #include <sys/time.h>
 #include <time.h>
+#if defined(__linux__)
+#include <linux/auxvec.h>
+#endif
 
 /***** arm *****/
 
@@ -57,6 +60,45 @@ orc_profile_stamp_xscale(void)
 }
 #endif
 
+#if defined(__linux__)
+static unsigned long
+orc_check_neon_proc_auxv (void)
+{
+  unsigned long flags = 0;
+  unsigned long aux[2];
+  ssize_t count;
+  int fd;
+
+  fd = open("/proc/self/auxv", O_RDONLY);
+  if (fd < 0) {
+    return 0;
+  }
+
+  while (1) {
+    count = read(fd, aux, sizeof(aux));
+    if (count < sizeof(aux)) {
+      break;
+    }
+
+    if (aux[0] == AT_HWCAP) {
+      //if (aux[1] & 64) flags |= ORC_TARGET_NEON_VFP;
+      //if (aux[1] & 512) flags |= ORC_TARGET_NEON_IWMMXT;
+      if (aux[1] & 4096) flags |= ORC_TARGET_NEON_NEON;
+      ORC_INFO("arm hwcap %08x", aux[1]);
+    } if (aux[0] == AT_PLATFORM) {
+      ORC_INFO("arm platform %s", (char *)aux[1]);
+    } else if (aux[0] == AT_NULL) {
+      break;
+    }
+  }
+
+  close(fd);
+
+  return flags;
+}
+#endif
+
+#ifdef unused
 static void
 orc_cpu_arm_getflags_cpuinfo (char *cpuinfo)
 {
@@ -70,7 +112,7 @@ orc_cpu_arm_getflags_cpuinfo (char *cpuinfo)
     return;
   }
 
-  flags = strsplit(cpuinfo_flags);
+  flags = strsplit(cpuinfo_flags, ' ');
   for (f = flags; *f; f++) {
 #if 0
     if (strcmp (*f, "edsp") == 0) {
@@ -117,10 +159,17 @@ get_proc_cpuinfo (void)
 
   return cpuinfo;
 }
+#endif
 
-void
-orc_cpu_detect_arch(void)
+unsigned long
+orc_arm_get_cpu_flags (void)
 {
+  unsigned long neon_flags = 0;
+
+#ifdef __linux__
+  neon_flags = orc_check_neon_proc_auxv ();
+#endif
+#ifdef unused
 #ifdef __linux__
   int arm_implementer = 0;
   char *cpuinfo;
@@ -159,6 +208,12 @@ orc_cpu_detect_arch(void)
   orc_cpu_arm_getflags_cpuinfo (cpuinfo);
   free (cpuinfo);
 #endif
+#endif
+  if (orc_compiler_flag_check ("-neon")) {
+    neon_flags &= ~ORC_TARGET_NEON_NEON;
+  }
+
+  return neon_flags;
 }
 #endif
 
