@@ -40,7 +40,6 @@ static CRITICAL_SECTION once_mutex;
 void
 _orc_once_init (void)
 {
-  InitializeCriticalSection (&once_mutex);
 }
 
 void
@@ -55,6 +54,15 @@ orc_once_mutex_unlock (void)
   LeaveCriticalSection (&once_mutex);
 }
 
+BOOL WINAPI
+DllMain (HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
+{
+  if (dwReason == DLL_PROCESS_ATTACH) {
+    InitializeCriticalSection (&once_mutex);
+  }
+  return TRUE;
+}
+
 #else
 
 void
