@@ -24,6 +24,7 @@ void orc_compiler_orc_arm_assemble (OrcCompiler *compiler);
 
 void orc_compiler_rewrite_vars (OrcCompiler *compiler);
 void orc_compiler_dump (OrcCompiler *compiler);
+void arm_add_strides (OrcCompiler *compiler);
 
 void
 orc_arm_emit_prologue (OrcCompiler *compiler)
@@ -115,7 +116,6 @@ orc_compiler_orc_arm_init (OrcCompiler *compiler)
   for(i=ORC_GP_REG_BASE;i<ORC_GP_REG_BASE+9;i++){
     compiler->valid_regs[i] = 1;
   }
-  compiler->valid_regs[ORC_ARM_A1] = 0;
   //compiler->valid_regs[ORC_ARM_SB] = 0;
   compiler->valid_regs[ORC_ARM_IP] = 0;
   compiler->valid_regs[ORC_ARM_SP] = 0;
@@ -129,24 +129,53 @@ orc_compiler_orc_arm_init (OrcCompiler *compiler)
     compiler->alloc_regs[i] = 0;
     compiler->used_regs[i] = 0;
   }
+  compiler->exec_reg = ORC_ARM_A1;
+  compiler->valid_regs[compiler->exec_reg] = 0;
+  compiler->gp_tmpreg = ORC_ARM_A2;
+  compiler->valid_regs[compiler->gp_tmpreg] = 0;
+  compiler->tmpreg = ORC_ARM_A3;
+  compiler->valid_regs[compiler->tmpreg] = 0;
 
   compiler->loop_shift = 0;
 }
 
 void
-orc_arm_load_constants (OrcCompiler *compiler)
+orc_arm_load_constants_outer (OrcCompiler *compiler)
 {
   int i;
   for(i=0;i<ORC_N_VARIABLES;i++){
     if (compiler->vars[i].name == NULL) continue;
     switch (compiler->vars[i].vartype) {
       case ORC_VAR_TYPE_CONST:
-        //orc_arm_emit_loadiw (compiler, compiler->vars[i].alloc,
-        //    (int)compiler->vars[i].value);
+        orc_arm_emit_load_imm (compiler, compiler->vars[i].alloc,
+            (int)compiler->vars[i].value);
         break;
       case ORC_VAR_TYPE_PARAM:
-        //orc_arm_emit_loadw (compiler, compiler->vars[i].alloc,
-        //    (int)ORC_STRUCT_OFFSET(OrcExecutor, params[i]), compiler->exec_reg);
+        ORC_PROGRAM_ERROR(compiler,"unimplemented");
+        return;
+        /* FIXME offset is too large */
+        //orc_arm_loadw (compiler, compiler->vars[i].alloc,
+        //    compiler->exec_reg,
+        //    (int)ORC_STRUCT_OFFSET(OrcExecutor, params[i]));
+        break;
+      case ORC_VAR_TYPE_SRC:
+      case ORC_VAR_TYPE_DEST:
+        break;
+      default:
+        break;
+    }
+  }
+}
+
+void
+orc_arm_load_constants_inner (OrcCompiler *compiler)
+{
+  int i;
+  for(i=0;i<ORC_N_VARIABLES;i++){
+    if (compiler->vars[i].name == NULL) continue;
+    switch (compiler->vars[i].vartype) {
+      case ORC_VAR_TYPE_CONST:
+      case ORC_VAR_TYPE_PARAM:
         break;
       case ORC_VAR_TYPE_SRC:
       case ORC_VAR_TYPE_DEST:
@@ -175,7 +204,8 @@ orc_arm_emit_load_src (OrcCompiler *compiler, OrcVariable *var)
     ptr_reg = var->ptr_register;
   }
   switch (var->size << compiler->loop_shift) {
-    //case 1:
+    case 1:
+      orc_arm_loadb (compiler, var->alloc, ptr_reg, 0);
       //orc_arm_emit_mov_memoffset_reg (compiler, 1, 0, ptr_reg, X86_ECX);
       //orc_arm_emit_mov_reg_arm (compiler, X86_ECX, var->alloc);
       break;
@@ -184,7 +214,8 @@ orc_arm_emit_load_src (OrcCompiler *compiler, OrcVariable *var)
       //orc_arm_emit_mov_memoffset_reg (compiler, 2, 0, ptr_reg, X86_ECX);
       //orc_arm_emit_mov_reg_arm (compiler, X86_ECX, var->alloc);
       break;
-    //case 4:
+    case 4:
+      orc_arm_loadl (compiler, var->alloc, ptr_reg, 0);
       //orc_arm_emit_mov_memoffset_arm (compiler, 4, 0, ptr_reg, var->alloc);
       break;
     //case 8:
@@ -194,7 +225,7 @@ orc_arm_emit_load_src (OrcCompiler *compiler, OrcVariable *var)
       //orc_arm_emit_mov_memoffset_arm (compiler, 16, 0, ptr_reg, var->alloc);
       break;
     default:
-      ORC_COMPILER_ERROR(compiler, "bad size %d\n", var->size << compiler->loop_shift);
+      ORC_COMPILER_ERROR(compiler, "bad size %d", var->size << compiler->loop_shift);
   }
 }
 
@@ -211,6 +242,7 @@ orc_arm_emit_store_dest (OrcCompiler *compiler, OrcVariable *var)
   }
   switch (var->size << compiler->loop_shift) {
     case 1:
+      orc_arm_storeb (compiler, ptr_reg, 0, var->alloc);
       //orc_arm_emit_mov_orc_arm_reg (compiler, var->alloc, X86_ECX);
       //orc_arm_emit_mov_reg_memoffset (compiler, 1, X86_ECX, 0, ptr_reg);
       break;
@@ -220,6 +252,7 @@ orc_arm_emit_store_dest (OrcCompiler *compiler, OrcVariable *var)
       //orc_arm_emit_mov_reg_memoffset (compiler, 2, X86_ECX, 0, ptr_reg);
       break;
     case 4:
+      orc_arm_storel (compiler, ptr_reg, 0, var->alloc);
       //orc_arm_emit_mov_orc_arm_memoffset (compiler, 4, var->alloc, 0, ptr_reg,
       //    var->is_aligned, var->is_uncached);
       break;
@@ -232,7 +265,7 @@ orc_arm_emit_store_dest (OrcCompiler *compiler, OrcVariable *var)
       //    var->is_aligned, var->is_uncached);
       break;
     default:
-      ORC_COMPILER_ERROR(compiler, "bad size %d\n", var->size << compiler->loop_shift);
+      ORC_COMPILER_ERROR(compiler, "bad size %d", var->size << compiler->loop_shift);
   }
 }
 
@@ -245,9 +278,27 @@ orc_compiler_orc_arm_assemble (OrcCompiler *compiler)
 
   orc_arm_emit_prologue (compiler);
 
+  orc_arm_load_constants_outer (compiler);
+
+  if (compiler->program->is_2d) {
+    if (compiler->program->constant_m > 0) {
+      orc_arm_emit_load_imm (compiler, ORC_ARM_A3, compiler->program->constant_m
+);    
+      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    } else {
+      orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A1]));
+      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    }
+    
+    orc_arm_emit_label (compiler, 8);
+  }
+
   orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
       (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
-  orc_arm_load_constants (compiler);
+  orc_arm_load_constants_inner (compiler);
 
   orc_arm_emit_label (compiler, 1);
 
@@ -261,6 +312,17 @@ orc_compiler_orc_arm_assemble (OrcCompiler *compiler)
   orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, 2);
   orc_arm_emit_label (compiler, 3);
 
+  if (compiler->program->is_2d) {
+    arm_add_strides (compiler);
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_A3, ORC_ARM_A3, 1, TRUE);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, 8);
+  }
+
   orc_arm_emit_epilogue (compiler);
 
   orc_arm_do_fixups (compiler);
@@ -298,6 +360,7 @@ orc_arm_emit_loop (OrcCompiler *compiler)
 
       switch (compiler->vars[insn->src_args[k]].vartype) {
         case ORC_VAR_TYPE_SRC:
+        case ORC_VAR_TYPE_DEST:
           orc_arm_emit_load_src (compiler, &compiler->vars[insn->src_args[k]]);
           break;
         case ORC_VAR_TYPE_CONST:
@@ -317,7 +380,7 @@ orc_arm_emit_loop (OrcCompiler *compiler)
       int dest = ORC_DEST_ARG (compiler, insn, 0);
 
       if (dest != src) {
-        orc_arm_emit_mov_r (compiler, ORC_ARM_COND_AL, 0, src, dest);
+        orc_arm_emit_mov_r (compiler, ORC_ARM_COND_AL, 0, dest, src);
       }
       rule->emit (compiler, rule->emit_user, insn);
     } else {
@@ -358,3 +421,36 @@ orc_arm_emit_loop (OrcCompiler *compiler)
   }
 }
 
+void
+arm_add_strides (OrcCompiler *compiler)
+{
+  int i;
+
+  for(i=0;i<ORC_N_VARIABLES;i++){
+    if (compiler->vars[i].name == NULL) continue;
+    switch (compiler->vars[i].vartype) {
+      case ORC_VAR_TYPE_CONST:
+        break;
+      case ORC_VAR_TYPE_PARAM:
+        break;
+      case ORC_VAR_TYPE_SRC:
+      case ORC_VAR_TYPE_DEST:
+        orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+            (int)ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]));
+        orc_arm_emit_load_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
+            (int)ORC_STRUCT_OFFSET(OrcExecutor, params[i]));
+        orc_arm_emit_add (compiler, ORC_ARM_A3, ORC_ARM_A3, ORC_ARM_A2);
+        orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+            (int)ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]));
+        break;
+      case ORC_VAR_TYPE_ACCUMULATOR:
+        break;
+      case ORC_VAR_TYPE_TEMP:
+        break;
+      default:
+        ORC_COMPILER_ERROR(compiler,"bad vartype");
+        break;
+    }
+  }
+}
+
