@@ -26,6 +26,52 @@ orc_compiler_c_init (OrcCompiler *compiler)
   compiler->loop_shift = 0;
 }
 
+const char *
+orc_target_c_get_typedefs (void)
+{
+  return
+    "#ifndef _ORC_INTEGER_TYPEDEFS_\n"
+    "#define _ORC_INTEGER_TYPEDEFS_\n"
+    "#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L\n"
+    "#include <stdint.h>\n"
+    "typedef int8_t orc_int8;\n"
+    "typedef int16_t orc_int16;\n"
+    "typedef int32_t orc_int32;\n"
+    "typedef int64_t orc_int64;\n"
+    "typedef uint8_t orc_uint8;\n"
+    "typedef uint16_t orc_uint16;\n"
+    "typedef uint32_t orc_uint32;\n"
+    "typedef uint64_t orc_uint64;\n"
+    "#elif defined(_MSC_VER)\n"
+    "typedef signed __int8 orc_int8;\n"
+    "typedef signed __int16 orc_int16;\n"
+    "typedef signed __int32 orc_int32;\n"
+    "typedef signed __int64 orc_int64;\n"
+    "typedef unsigned __int8 orc_uint8;\n"
+    "typedef unsigned __int16 orc_uint16;\n"
+    "typedef unsigned __int32 orc_uint32;\n"
+    "typedef unsigned __int64 orc_uint64;\n"
+    "#else\n"
+    "#include <limits.h>\n"
+    "typedef signed char orc_int8;\n"
+    "typedef short orc_int16;\n"
+    "typedef int orc_int32;\n"
+    "typedef unsigned char orc_uint8;\n"
+    "typedef unsigned short orc_uint16;\n"
+    "typedef unsigned int orc_uint32;\n"
+    "#if INT_MAX == LONG_MAX\n"
+    "typedef long long orc_int64;\n"
+    "typedef unsigned long long orc_uint64;\n"
+    "#else\n"
+    "typedef long orc_int64;\n"
+    "typedef unsigned long orc_uint64;\n"
+    "#endif\n"
+    "#endif\n"
+    "typedef union { orc_int32 i; float f; } orc_union32;\n"
+    "typedef union { orc_int64 i; double f; } orc_union64;\n"
+    "#endif\n";
+}
+
 const char *
 orc_target_c_get_asm_preamble (void)
 {
@@ -56,9 +102,6 @@ orc_target_c_get_asm_preamble (void)
     "#define ORC_SWAP_W(x) ((((x)&0xff)<<8) | (((x)&0xff00)>>8))\n"
     "#define ORC_SWAP_L(x) ((((x)&0xff)<<24) | (((x)&0xff00)<<8) | (((x)&0xff0000)>>8) | (((x)&0xff000000)>>24))\n"
     "#define ORC_PTR_OFFSET(ptr,offset) ((void *)(((unsigned char *)(ptr)) + (offset)))\n"
-    "#define ORC_AS_FLOAT(x) (((union { int i; float f; } *)(&x))->f)\n"
-    "typedef union { int32_t i; float f; } orc_union32;\n"
-    "typedef union { int64_t i; double f; } orc_union64;\n"
     "/* end Orc C target preamble */\n\n";
 }
 
@@ -199,22 +242,17 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
       case ORC_VAR_TYPE_PARAM:
         c_get_name (varname, compiler, i);
         if (!(compiler->target_flags & ORC_TARGET_C_NOEXEC)) {
-          if (var->size >= 4) {
-            ORC_ASM_CODE(compiler,"  const %s %s = *(%s *)(ex->params + %d);\n",
-                c_get_type_name (var->size), varname,
-                c_get_type_name (var->size), i);
+          if (var->is_float_param) {
+            ORC_ASM_CODE(compiler,"  const float %s = ((orc_union32 *)(ex->params+%d))->f;\n",
+                varname, i);
           } else {
-            ORC_ASM_CODE(compiler,"  const %s %s = ex->params[%d];\n",
-                c_get_type_name (var->size), varname, i);
+            ORC_ASM_CODE(compiler,"  const int %s = ex->params[%d];\n",
+                varname, i);
           }
         } else {
-          if (var->size >= 4) {
-            ORC_ASM_CODE(compiler,"  const %s %s = { %s };\n",
-                c_get_type_name (var->size), varname, varnames[i]);
-          } else {
-            ORC_ASM_CODE(compiler,"  const %s %s = %s;\n",
-                c_get_type_name (var->size), varname, varnames[i]);
-          }
+          ORC_ASM_CODE(compiler,"  const %s %s = %s;\n",
+              var->is_float_param ? "float" : "int",
+              varname, varnames[i]);
         }
         break;
       default:
@@ -237,27 +275,9 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
             char s1[20], s2[20];
             get_varname(s1, compiler, i);
             get_varname_stride(s2, compiler, i);
-            switch (var->sampling_type) {
-              case ORC_SAMPLE_REGULAR:
-                ORC_ASM_CODE(compiler,
-                    "    ptr%d = ORC_PTR_OFFSET(%s, %s * j);\n",
-                    i, s1, s2);
-                break;
-              case ORC_SAMPLE_TRANSPOSED:
-                ORC_ASM_CODE(compiler,
-                    "    ptr%d = ORC_PTR_OFFSET(%s, %d * j);\n",
-                    i, s1, var->size);
-                break;
-              case ORC_SAMPLE_NEAREST:
-              case ORC_SAMPLE_BILINEAR:
-              case ORC_SAMPLE_FOUR_TAP:
-                ORC_ASM_CODE(compiler,
-                    "    ptr%d = ORC_PTR_OFFSET(%s, %s * j);\n",
-                    i, s1, s2);
-                break;
-              default:
-                ORC_COMPILER_ERROR(compiler, "eeek");
-            }
+            ORC_ASM_CODE(compiler,
+                "    ptr%d = ORC_PTR_OFFSET(%s, %s * j);\n",
+                i, s1, s2);
           }
           break;
         case ORC_VAR_TYPE_DEST:
@@ -325,7 +345,8 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
       ORC_ASM_CODE(compiler,"%*s", prefix, "");
       rule->emit (compiler, rule->emit_user, insn);
     } else {
-      ORC_COMPILER_ERROR(compiler, "No rule for: %s\n", opcode->name);
+      ORC_COMPILER_ERROR(compiler, "No rule for: %s on target %s", opcode->name,
+          compiler->target->name);
       compiler->error = TRUE;
     }
   }
@@ -408,7 +429,6 @@ c_get_name_int (char *name, OrcCompiler *p, int var)
 {
   switch (p->vars[var].vartype) {
     case ORC_VAR_TYPE_CONST:
-    case ORC_VAR_TYPE_PARAM:
     case ORC_VAR_TYPE_TEMP:
     case ORC_VAR_TYPE_ACCUMULATOR:
     case ORC_VAR_TYPE_SRC:
@@ -419,6 +439,9 @@ c_get_name_int (char *name, OrcCompiler *p, int var)
         sprintf(name, "var%d", var);
       }
       break;
+    case ORC_VAR_TYPE_PARAM:
+      sprintf(name, "var%d", var);
+      break;
     default:
       ORC_COMPILER_ERROR(p, "bad vartype");
       sprintf(name, "ERROR");
@@ -431,13 +454,15 @@ c_get_name_float (char *name, OrcCompiler *p, int var)
 {
   switch (p->vars[var].vartype) {
     case ORC_VAR_TYPE_CONST:
-    case ORC_VAR_TYPE_PARAM:
     case ORC_VAR_TYPE_TEMP:
     case ORC_VAR_TYPE_ACCUMULATOR:
     case ORC_VAR_TYPE_SRC:
     case ORC_VAR_TYPE_DEST:
       sprintf(name, "var%d.f", var);
       break;
+    case ORC_VAR_TYPE_PARAM:
+      sprintf(name, "var%d", var);
+      break;
     default:
       ORC_COMPILER_ERROR(p, "bad vartype");
       sprintf(name, "ERROR");
@@ -450,9 +475,9 @@ c_get_type_name (int size)
 {
   switch (size) {
     case 1:
-      return "int8_t";
+      return "orc_int8";
     case 2:
-      return "int16_t";
+      return "orc_int16";
     case 4:
       return "orc_union32";
     case 8:
@@ -637,7 +662,7 @@ c_rule_accsadubl (OrcCompiler *p, void *user, OrcInstruction *insn)
   c_get_name_int (src2, p, insn->src_args[1]);
 
   ORC_ASM_CODE(p,
-      "    %s = %s + ORC_ABS((int32_t)(uint8_t)%s - (int32_t)(uint8_t)%s);\n",
+      "    %s = %s + ORC_ABS((orc_int32)(orc_uint8)%s - (orc_int32)(orc_uint8)%s);\n",
       dest, dest, src1, src2);
 }
 
