@@ -480,7 +480,8 @@ emit_loop (OrcCompiler *compiler, int prefix)
       ORC_ASM_CODE(compiler,"%*s", prefix, "");
       rule->emit (compiler, rule->emit_user, insn);
     } else {
-      ORC_COMPILER_ERROR(compiler, "No rule for: %s\n", opcode->name);
+      ORC_COMPILER_ERROR(compiler, "No rule for: %s on target %s",
+          opcode->name, compiler->target->name);
       compiler->error = TRUE;
     }
   }
