@@ -5,7 +5,6 @@
 #include <string.h>
 #include <stdlib.h>
 
-#include <unistd.h>
 #include <sys/types.h>
 
 #include <orc/orcprogram.h>
@@ -17,7 +16,7 @@
 
 #define SIZE 65536
 
-void orc_neon_emit_loop (OrcCompiler *compiler);
+void orc_neon_emit_loop (OrcCompiler *compiler, int unroll_index);
 
 void orc_compiler_neon_register_rules (OrcTarget *target);
 unsigned int orc_compiler_neon_get_default_flags (void);
@@ -28,6 +27,7 @@ void orc_compiler_neon_assemble (OrcCompiler *compiler);
 void orc_compiler_rewrite_vars (OrcCompiler *compiler);
 void orc_compiler_dump (OrcCompiler *compiler);
 void orc_neon_save_accumulators (OrcCompiler *compiler);
+void neon_add_strides (OrcCompiler *compiler);
 
 
 void
@@ -102,6 +102,13 @@ static OrcTarget neon_target = {
 void
 orc_neon_init (void)
 {
+#if defined(HAVE_ARM)
+  if (!(orc_arm_get_cpu_flags () & ORC_TARGET_NEON_NEON)) {
+    ORC_INFO("marking neon backend non-executable");
+    neon_target.executable = FALSE;
+  }
+#endif
+
   orc_target_register (&neon_target);
 
   orc_compiler_neon_register_rules (&neon_target);
@@ -110,7 +117,7 @@ orc_neon_init (void)
 unsigned int
 orc_compiler_neon_get_default_flags (void)
 {
-  return 0;
+  return ORC_TARGET_NEON_NEON;
 }
 
 void
@@ -168,16 +175,16 @@ orc_compiler_neon_init (OrcCompiler *compiler)
 
   switch (orc_program_get_max_array_size (compiler->program)) {
     case 1:
-      loop_shift = 3;
+      loop_shift = 4;
       break;
     case 2:
-      loop_shift = 2;
+      loop_shift = 3;
       break;
     case 4:
-      loop_shift = 1;
+      loop_shift = 2;
       break;
     case 8:
-      loop_shift = 0;
+      loop_shift = 1;
       break;
     default:
       ORC_ERROR("unhandled max var size %d",
@@ -213,11 +220,14 @@ orc_compiler_neon_init (OrcCompiler *compiler)
     compiler->loop_shift = loop_shift;
   }
 
-  compiler->need_mask_regs = TRUE;
+  compiler->unroll_shift = 1;
+  if (0) {
+    compiler->need_mask_regs = TRUE;
+  }
 }
 
 void
-orc_neon_load_constants (OrcCompiler *compiler)
+orc_neon_load_constants_outer (OrcCompiler *compiler)
 {
   int i;
   for(i=0;i<ORC_N_VARIABLES;i++){
@@ -250,13 +260,39 @@ orc_neon_load_constants (OrcCompiler *compiler)
         }
         break;
       case ORC_VAR_TYPE_SRC:
+      case ORC_VAR_TYPE_DEST:
+        break;
+      case ORC_VAR_TYPE_ACCUMULATOR:
+        orc_neon_emit_loadil (compiler, compiler->vars[i].alloc, 0);
+        break;
+      case ORC_VAR_TYPE_TEMP:
+        break;
+      default:
+        ORC_PROGRAM_ERROR(compiler,"bad vartype");
+        break;
+    }
+  }
+}
+
+void
+orc_neon_load_constants_inner (OrcCompiler *compiler)
+{
+  int i;
+  for(i=0;i<ORC_N_VARIABLES;i++){
+    if (compiler->vars[i].name == NULL) continue;
+
+    switch (compiler->vars[i].vartype) {
+      case ORC_VAR_TYPE_CONST:
+        break;
+      case ORC_VAR_TYPE_PARAM:
+        break;
+      case ORC_VAR_TYPE_SRC:
       case ORC_VAR_TYPE_DEST:
         orc_arm_emit_load_reg (compiler, 
             compiler->vars[i].ptr_register,
             compiler->exec_reg, ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]));
         break;
       case ORC_VAR_TYPE_ACCUMULATOR:
-        orc_neon_emit_loadil (compiler, compiler->vars[i].alloc, 0);
         break;
       case ORC_VAR_TYPE_TEMP:
         break;
@@ -331,14 +367,14 @@ orc_neon_load_alignment_masks (OrcCompiler *compiler)
             code |= ((var->mask_alloc>>4)&0x1) << 22;
             orc_arm_emit (compiler, code);
 
-            orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, 8+b);
+            orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, 9+b);
             for(j=0;j<8;j++){
               ORC_ASM_CODE(compiler, "  .word 0x%02x%02x%02x%02x\n", j+3, j+2, j+1, j+0);
               orc_arm_emit (compiler, ((j+0)<<0) | ((j+1)<<8) | ((j+2)<<16) | ((j+3)<<24));
               ORC_ASM_CODE(compiler, "  .word 0x%02x%02x%02x%02x\n", j+7, j+6, j+5, j+4);
               orc_arm_emit (compiler, ((j+4)<<0) | ((j+5)<<8) | ((j+6)<<16) | ((j+7)<<24));
             }
-            orc_arm_emit_label (compiler, 8+b);
+            orc_arm_emit_label (compiler, 9+b);
             b++;
 
           }
@@ -411,7 +447,7 @@ orc_neon_restore_unalignment (OrcCompiler *compiler)
           orc_arm_emit_add (compiler, var->ptr_register, var->ptr_register,
               var->ptr_offset);
           orc_arm_emit_sub_imm (compiler, var->ptr_register, var->ptr_register,
-              size);
+              size, FALSE);
         }
         break;
       case ORC_VAR_TYPE_DEST:
@@ -429,7 +465,7 @@ orc_neon_restore_unalignment (OrcCompiler *compiler)
 }
 
 void
-orc_neon_emit_load_src (OrcCompiler *compiler, OrcVariable *var)
+orc_neon_emit_load_src (OrcCompiler *compiler, OrcVariable *var, int unroll_index)
 {
   int ptr_reg;
   int update;
@@ -465,6 +501,19 @@ orc_neon_emit_load_src (OrcCompiler *compiler, OrcVariable *var)
     default:
       ORC_ERROR("bad size");
   }
+  
+  if (unroll_index == 0) {
+  switch (compiler->size_region) {
+    case 0:
+    case 1:
+      orc_neon_preload (compiler, var, FALSE, 208);
+      break;
+    case 2:
+    case 3:
+      orc_neon_preload (compiler, var, FALSE, 208);
+      break;
+  }
+  }
 }
 
 void
@@ -494,6 +543,24 @@ orc_neon_emit_store_dest (OrcCompiler *compiler, OrcVariable *var)
     default:
       ORC_ERROR("bad size");
   }
+
+  switch (compiler->size_region) {
+    case 0:
+      break;
+    case 1:
+      /* assume hot cache, see below */
+      break;
+    case 2:
+      /* This is only useful for cold cache and for memset-like operations,
+         which isn't the usual case, thus it's disabled. */
+#if 0
+      orc_neon_preload (compiler, var, FALSE, 208);
+#endif
+      break;
+    case 3:
+      /* none */
+      break;
+  }
 }
 
 static int
@@ -525,28 +592,98 @@ get_align_var (OrcCompiler *compiler)
   return -1;
 }
 
+enum {
+  LABEL_ONE_REGION = 1,
+  LABEL_ONE_REGION_AFTER,
+  LABEL_REGION0_LOOP,
+  LABEL_REGION0_SKIP,
+  LABEL_REGION1_LOOP,
+  LABEL_REGION1_SKIP,
+  LABEL_REGION2_LOOP_SMALL,
+  LABEL_REGION2_LOOP_MEDIUM,
+  LABEL_REGION2_LOOP_LARGE,
+  LABEL_REGION2_SMALL,
+  LABEL_REGION2_MEDIUM,
+  LABEL_REGION2_SKIP,
+  LABEL_REGION3_LOOP,
+  LABEL_REGION3_SKIP,
+  LABEL_OUTER_LOOP,
+  LABEL_OUTER_LOOP_SKIP,
+  LABEL_L1L2_AFTER,
+};
+
 void
 orc_compiler_neon_assemble (OrcCompiler *compiler)
 {
   int align_var;
   int align_shift;
   int var_size_shift;
+  int i;
   
   align_var = get_align_var (compiler);
   var_size_shift = get_shift (compiler->vars[align_var].size);
-  align_shift = var_size_shift + compiler->loop_shift;
+  align_shift = 4;
 
   compiler->vars[align_var].is_aligned = FALSE;
 
   orc_neon_emit_prologue (compiler);
 
+  orc_neon_load_constants_outer (compiler);
+
+  if (compiler->program->is_2d) {
+    if (compiler->program->constant_m > 0) {
+      orc_arm_emit_load_imm (compiler, ORC_ARM_A3, compiler->program->constant_m);
+      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    } else {
+      orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A1]));
+      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    }
+
+    orc_arm_emit_label (compiler, LABEL_OUTER_LOOP);
+  }
+
+#define ORC_NEON_ALIGNED_DEST_CUTOFF 64
+
   if (compiler->loop_shift > 0) {
-    int align_shift = 3;
-
-    orc_arm_emit_load_imm (compiler, ORC_ARM_IP, 1<<align_shift);
-
     orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
+    orc_arm_emit_cmp_imm (compiler, ORC_ARM_A3, ORC_NEON_ALIGNED_DEST_CUTOFF);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_GT, LABEL_REGION0_SKIP);
+
+    orc_arm_emit_asr_imm (compiler, ORC_ARM_A2, ORC_ARM_A3,
+        compiler->loop_shift);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+
+    orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A3,
+        (1<<compiler->loop_shift)-1);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
+
+    orc_neon_load_constants_inner (compiler);
+    orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
+
+    compiler->size_region = 0;
+    orc_arm_emit_label (compiler, LABEL_REGION0_LOOP);
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+    orc_neon_emit_loop (compiler, -1);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION0_LOOP);
+    
+
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
+
+    orc_arm_emit_label (compiler, LABEL_REGION0_SKIP);
+  }
+
+  if (compiler->loop_shift > 0) {
+    orc_arm_emit_load_imm (compiler, ORC_ARM_IP, 1<<align_shift);
+
     orc_arm_emit_load_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,arrays[align_var]));
     orc_arm_emit_sub (compiler, ORC_ARM_IP, ORC_ARM_IP, ORC_ARM_A2);
@@ -556,23 +693,27 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
       orc_arm_emit_asr_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, var_size_shift);
     }
 
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
     orc_arm_emit_cmp (compiler, ORC_ARM_A3, ORC_ARM_IP);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_LE, 6);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_LE, LABEL_ONE_REGION);
 
     orc_arm_emit_store_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
     orc_arm_emit_sub (compiler, ORC_ARM_A2, ORC_ARM_A3, ORC_ARM_IP);
 
-    orc_arm_emit_asr_imm (compiler, ORC_ARM_A3, ORC_ARM_A2, compiler->loop_shift);
+    orc_arm_emit_asr_imm (compiler, ORC_ARM_A3, ORC_ARM_A2,
+        compiler->loop_shift + compiler->unroll_shift);
     orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
 
-    orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A2, (1<<compiler->loop_shift)-1);
+    orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A2,
+        (1<<(compiler->loop_shift + compiler->unroll_shift))-1);
     orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
 
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, 7);
-    orc_arm_emit_label (compiler, 6);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_ONE_REGION_AFTER);
+    orc_arm_emit_label (compiler, LABEL_ONE_REGION);
 
     orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
@@ -583,10 +724,10 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
     orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
 
-    orc_arm_emit_label (compiler, 7);
+    orc_arm_emit_label (compiler, LABEL_ONE_REGION_AFTER);
   }
 
-  orc_neon_load_constants (compiler);
+  orc_neon_load_constants_inner (compiler);
 
   if (compiler->loop_shift > 0) {
     int save_loop_shift = compiler->loop_shift;
@@ -596,14 +737,13 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
 
     orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, 1);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION1_SKIP);
 
-    orc_arm_emit_label (compiler, 0);
-    orc_neon_emit_loop (compiler);
-    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1);
-    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, 0);
-    orc_arm_emit_label (compiler, 1);
+    orc_arm_emit_label (compiler, LABEL_REGION1_LOOP);
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+    orc_neon_emit_loop (compiler, -1);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION1_LOOP);
+    orc_arm_emit_label (compiler, LABEL_REGION1_SKIP);
 
     compiler->loop_shift = save_loop_shift;
     compiler->vars[align_var].is_aligned = TRUE;
@@ -618,23 +758,64 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
   }
 
   orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, 3);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
 
-  orc_neon_load_alignment_masks (compiler);
+  if (0) {
+    /* Disable alignment masks for now.  It can easily take all available
+       registers. */
+    orc_neon_load_alignment_masks (compiler);
+  }
 
-  orc_arm_emit_label (compiler, 2);
-  orc_neon_emit_loop (compiler);
-  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1);
-  orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, 2);
+  orc_arm_emit_asr_imm (compiler, compiler->gp_tmpreg, ORC_ARM_IP,
+      17 + var_size_shift - compiler->loop_shift - compiler->unroll_shift);
+  orc_arm_emit_cmp_imm (compiler, compiler->gp_tmpreg, 0);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_MEDIUM);
 
-  orc_neon_restore_unalignment (compiler);
+  /* N is larger than L2 cache size */
+  compiler->size_region = 3;
+  orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_LARGE);
+  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+  for(i=0;i<(1<<compiler->unroll_shift);i++){
+    orc_neon_emit_loop (compiler, i);
+  }
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_LARGE);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
 
-  orc_arm_emit_label (compiler, 3);
+  orc_arm_emit_label (compiler, LABEL_REGION2_MEDIUM);
+  orc_arm_emit_asr_imm (compiler, compiler->gp_tmpreg, ORC_ARM_IP,
+      13 + var_size_shift - compiler->loop_shift - compiler->unroll_shift);
+  orc_arm_emit_cmp_imm (compiler, compiler->gp_tmpreg, 0);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SMALL);
 
+  /* N is smaller than L2 cache size */
+  compiler->size_region = 2;
+  orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_MEDIUM);
+  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+  for(i=0;i<(1<<compiler->unroll_shift);i++){
+    orc_neon_emit_loop (compiler, i);
+  }
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_MEDIUM);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
+
+  orc_arm_emit_label (compiler, LABEL_REGION2_SMALL);
+  /* N is smaller than L2 cache size */
+  compiler->size_region = 1;
+  orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_SMALL);
+  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+  for(i=0;i<(1<<compiler->unroll_shift);i++){
+    orc_neon_emit_loop (compiler, i);
+  }
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_SMALL);
+
+  if (0) {
+    orc_neon_restore_unalignment (compiler);
+  }
+
+  orc_arm_emit_label (compiler, LABEL_REGION2_SKIP);
 
   if (compiler->loop_shift > 0) {
     int save_loop_shift = compiler->loop_shift;
+
     compiler->loop_shift = 0;
 
     compiler->vars[align_var].is_aligned = FALSE;
@@ -643,18 +824,28 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
 
     orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, 5);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION3_SKIP);
 
-    orc_arm_emit_label (compiler, 4);
-    orc_neon_emit_loop (compiler);
-    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1);
-    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, 4);
-    orc_arm_emit_label (compiler, 5);
+    orc_arm_emit_label (compiler, LABEL_REGION3_LOOP);
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+    orc_neon_emit_loop (compiler, -1);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION3_LOOP);
+    orc_arm_emit_label (compiler, LABEL_REGION3_SKIP);
 
     compiler->loop_shift = save_loop_shift;
   }
 
+  if (compiler->program->is_2d) {
+    neon_add_strides (compiler);
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_A3, ORC_ARM_A3, 1, TRUE);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_OUTER_LOOP);
+  }
+
   orc_neon_save_accumulators (compiler);
 
   orc_neon_emit_epilogue (compiler);
@@ -667,7 +858,7 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
 }
 
 void
-orc_neon_emit_loop (OrcCompiler *compiler)
+orc_neon_emit_loop (OrcCompiler *compiler, int unroll_index)
 {
   int j;
   int k;
@@ -675,6 +866,7 @@ orc_neon_emit_loop (OrcCompiler *compiler)
   OrcStaticOpcode *opcode;
   OrcRule *rule;
 
+  orc_compiler_append_code(compiler,"# LOOP shift %d\n", compiler->loop_shift);
   for(j=0;j<compiler->n_insns;j++){
     compiler->insn_index = j;
     insn = compiler->insns + j;
@@ -700,7 +892,7 @@ orc_neon_emit_loop (OrcCompiler *compiler)
       switch (compiler->vars[insn->src_args[k]].vartype) {
         case ORC_VAR_TYPE_SRC:
         case ORC_VAR_TYPE_DEST:
-          orc_neon_emit_load_src (compiler, &compiler->vars[insn->src_args[k]]);
+          orc_neon_emit_load_src (compiler, &compiler->vars[insn->src_args[k]], unroll_index);
           break;
         case ORC_VAR_TYPE_CONST:
           break;
@@ -851,3 +1043,36 @@ orc_neon_save_accumulators (OrcCompiler *compiler)
   }
 }
 
+void
+neon_add_strides (OrcCompiler *compiler)
+{
+  int i;
+
+  for(i=0;i<ORC_N_VARIABLES;i++){
+    if (compiler->vars[i].name == NULL) continue;
+    switch (compiler->vars[i].vartype) {
+      case ORC_VAR_TYPE_CONST:
+        break;
+      case ORC_VAR_TYPE_PARAM:
+        break;
+      case ORC_VAR_TYPE_SRC:
+      case ORC_VAR_TYPE_DEST:
+        orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+            (int)ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]));
+        orc_arm_emit_load_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
+            (int)ORC_STRUCT_OFFSET(OrcExecutor, params[i]));
+        orc_arm_emit_add (compiler, ORC_ARM_A3, ORC_ARM_A3, ORC_ARM_A2);
+        orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+            (int)ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]));
+        break;
+      case ORC_VAR_TYPE_ACCUMULATOR:
+        break;
+      case ORC_VAR_TYPE_TEMP:
+        break;
+      default:
+        ORC_COMPILER_ERROR(compiler,"bad vartype");
+        break;
+    }
+  }
+}
+
