@@ -28,6 +28,7 @@ void orc_compiler_sse_register_rules (OrcTarget *target);
 void orc_compiler_rewrite_vars (OrcCompiler *compiler);
 void orc_compiler_dump (OrcCompiler *compiler);
 void sse_load_constant (OrcCompiler *compiler, int reg, int size, int value);
+static const char * sse_get_flag_name (int shift);
 
 static OrcTarget sse_target = {
   "sse",
@@ -43,14 +44,26 @@ static OrcTarget sse_target = {
   { { 0 } },
   0,
   NULL,
-  sse_load_constant
-
+  sse_load_constant,
+  sse_get_flag_name
 };
 
 
+
 void
 orc_sse_init (void)
 {
+#if defined(HAVE_AMD64) || defined(HAVE_I386)
+  /* initializes cache information */
+  orc_sse_get_cpu_flags ();
+#endif
+
+#if defined(HAVE_I386)
+  if (!(orc_sse_get_cpu_flags () & ORC_TARGET_SSE_SSE2)) {
+    sse_target.executable = FALSE;
+  }
+#endif
+
   orc_target_register (&sse_target);
 
   orc_compiler_sse_register_rules (&sse_target);
@@ -64,7 +77,9 @@ orc_compiler_sse_get_default_flags (void)
 #ifdef HAVE_AMD64
   flags |= ORC_TARGET_SSE_64BIT;
 #endif
-  flags &= ~ORC_TARGET_SSE_FRAME_POINTER;
+  if (_orc_compiler_flag_debug) {
+    flags |= ORC_TARGET_SSE_FRAME_POINTER;
+  }
   
 #if defined(HAVE_AMD64) || defined(HAVE_I386)
   flags |= orc_sse_get_cpu_flags ();
@@ -77,6 +92,21 @@ orc_compiler_sse_get_default_flags (void)
   return flags;
 }
 
+static const char *
+sse_get_flag_name (int shift)
+{
+  static const char *flags[] = {
+    "sse2", "sse3", "ssse3", "sse41", "sse42", "sse4a", "sse5",
+    "frame_pointer", "short_jumps", "64bit"
+  };
+
+  if (shift >= 0 && shift < sizeof(flags)/sizeof(flags[0])) {
+    return flags[shift];
+  }
+
+  return NULL;
+}
+
 void
 orc_compiler_sse_init (OrcCompiler *compiler)
 {
@@ -164,8 +194,8 @@ orc_compiler_sse_init (OrcCompiler *compiler)
       break;
   }
 
-  compiler->unroll_shift = 0;
-  //compiler->alloc_loop_counter = TRUE;
+  compiler->unroll_shift = 1;
+  compiler->alloc_loop_counter = TRUE;
 
 }
 
@@ -183,7 +213,7 @@ sse_save_accumulators (OrcCompiler *compiler)
       case ORC_VAR_TYPE_ACCUMULATOR:
         src = compiler->vars[i].alloc;
 
-        orc_sse_emit_pshufd (compiler, 0xee, src, compiler->tmpreg);
+        orc_sse_emit_pshufd (compiler, ORC_SSE_SHUF(3,2,3,2), src, compiler->tmpreg);
 
         if (compiler->vars[i].size == 2) {
           orc_sse_emit_660f (compiler, "paddw", 0xfd, compiler->tmpreg, src);
@@ -191,7 +221,7 @@ sse_save_accumulators (OrcCompiler *compiler)
           orc_sse_emit_660f (compiler, "paddd", 0xfe, compiler->tmpreg, src);
         }
 
-        orc_sse_emit_pshufd (compiler, 0x55, src, compiler->tmpreg);
+        orc_sse_emit_pshufd (compiler, ORC_SSE_SHUF(1,1,1,1), src, compiler->tmpreg);
 
         if (compiler->vars[i].size == 2) {
           orc_sse_emit_660f (compiler, "paddw", 0xfd, compiler->tmpreg, src);
@@ -200,7 +230,7 @@ sse_save_accumulators (OrcCompiler *compiler)
         }
 
         if (compiler->vars[i].size == 2) {
-          orc_sse_emit_pshuflw (compiler, 0x55, src, compiler->tmpreg);
+          orc_sse_emit_pshuflw (compiler, ORC_SSE_SHUF(1,1,1,1), src, compiler->tmpreg);
 
           orc_sse_emit_660f (compiler, "paddw", 0xfd, compiler->tmpreg, src);
         }
@@ -472,7 +502,7 @@ get_shift (int size)
 }
 
 
-void
+static void
 orc_emit_split_n_regions (OrcCompiler *compiler)
 {
   int align_var;
@@ -565,6 +595,8 @@ orc_compiler_sse_assemble (OrcCompiler *compiler)
 
   orc_x86_emit_prologue (compiler);
 
+  orc_sse_set_mxcsr (compiler);
+
   sse_load_constants_outer (compiler);
 
   if (compiler->program->is_2d) {
@@ -675,7 +707,7 @@ orc_compiler_sse_assemble (OrcCompiler *compiler)
           (ui==ui_max-1) << (compiler->loop_shift + compiler->unroll_shift));
     }
     if (compiler->loop_counter != ORC_REG_INVALID) {
-      orc_x86_emit_add_imm_reg (compiler, 4, -1, compiler->loop_counter);
+      orc_x86_emit_add_imm_reg (compiler, 4, -1, compiler->loop_counter, TRUE);
     } else {
       orc_x86_emit_dec_memoffset (compiler, 4,
           (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2),
@@ -718,6 +750,7 @@ orc_compiler_sse_assemble (OrcCompiler *compiler)
 
   sse_save_accumulators (compiler);
 
+  orc_sse_restore_mxcsr (compiler);
   orc_x86_emit_epilogue (compiler);
 
   orc_x86_do_fixups (compiler);
@@ -810,7 +843,7 @@ orc_sse_emit_loop (OrcCompiler *compiler, int offset, int update)
         if (compiler->vars[k].ptr_register) {
           orc_x86_emit_add_imm_reg (compiler, compiler->is_64bit ? 8 : 4,
               compiler->vars[k].size * update,
-              compiler->vars[k].ptr_register);
+              compiler->vars[k].ptr_register, FALSE);
         } else {
           orc_x86_emit_add_imm_memoffset (compiler, compiler->is_64bit ? 8 : 4,
               compiler->vars[k].size * update,
