@@ -35,23 +35,121 @@ arm_rule_ ## opcode (OrcCompiler *p, void *user, OrcInstruction *insn) \
 }
 
 /* multiplies */
-#define orc_arm_smulxy(cond,x,y,Rd,Rm,Rs) (0x01600080|((cond)<<28)|((Rd)<<16)|((Rs)<<8)|((y)<<6)|((x)<<5)|(Rm))
-#define orc_arm_emit_smulbb(p,cond,Rd,Rm,Rs) orc_arm_emit (p, orc_arm_smulxy (cond,0,0,Rd,Rm,Rs))
-#define orc_arm_emit_smulbt(p,cond,Rd,Rm,Rs) orc_arm_emit (p, orc_arm_smulxy (cond,0,1,Rd,Rm,Rs))
-#define orc_arm_emit_smultb(p,cond,Rd,Rm,Rs) orc_arm_emit (p, orc_arm_smulxy (cond,1,0,Rd,Rm,Rs))
-#define orc_arm_emit_smultt(p,cond,Rd,Rm,Rs) orc_arm_emit (p, orc_arm_smulxy (cond,1,1,Rd,Rm,Rs))
+#define orc_arm_smulxy(cond,x,y,Rd,Rm,Rs) (0x01600080|((cond)<<28)|(((Rd)&15)<<16)|(((Rs)&15)<<8)|((y)<<6)|((x)<<5)|((Rm)&15))
+#define orc_arm_emit_smulbb(p,cond,Rd,Rm,Rs) do { \
+  ORC_ASM_CODE (p, "smulbb %s, %s, %s\n", orc_arm_reg_name (Rd), \
+      orc_arm_reg_name (Rm), orc_arm_reg_name(Rs)); \
+  orc_arm_emit (p, orc_arm_smulxy (cond,0,0,Rd,Rm,Rs)); \
+} while (0)
+#define orc_arm_emit_smulbt(p,cond,Rd,Rm,Rs) do { \
+  ORC_ASM_CODE (p, "smulbt %s, %s, %s\n", orc_arm_reg_name (Rd), \
+      orc_arm_reg_name (Rm), orc_arm_reg_name(Rs)); \
+  orc_arm_emit (p, orc_arm_smulxy (cond,0,1,Rd,Rm,Rs)); \
+} while (0)
+#define orc_arm_emit_smultb(p,cond,Rd,Rm,Rs) do { \
+  ORC_ASM_CODE (p, "smultb %s, %s, %s\n", orc_arm_reg_name (Rd), \
+      orc_arm_reg_name (Rm), orc_arm_reg_name(Rs)); \
+  orc_arm_emit (p, orc_arm_smulxy (cond,1,0,Rd,Rm,Rs)); \
+} while (0)
+#define orc_arm_emit_smultt(p,cond,Rd,Rm,Rs) do { \
+  ORC_ASM_CODE (p, "smultt %s, %s, %s\n", orc_arm_reg_name (Rd), \
+      orc_arm_reg_name (Rm), orc_arm_reg_name(Rs)); \
+  orc_arm_emit (p, orc_arm_smulxy (cond,1,1,Rd,Rm,Rs)); \
+} while (0)
 
-#define orc_arm_mul(cond,S,Rd,Rm,Rs) (0x00000090|((cond)<<28)|((S)<<20)|((Rd)<<16)|((Rs)<<8)|(Rm))
-#define orc_arm_emit_mul(p,cond,S,Rd,Rm,Rs) orc_arm_emit (p, orc_arm_mul (cond,S,Rd,Rm,Rs))
+#define orc_arm_mul(cond,S,Rd,Rm,Rs) (0x00000090|((cond)<<28)|((S)<<20)|(((Rd)&15)<<16)|(((Rs)&15)<<8)|((Rm)&15))
+#define orc_arm_emit_mul(p,cond,S,Rd,Rm,Rs) do { \
+  ORC_ASM_CODE (p, "mul %s, %s, %s\n", orc_arm_reg_name (Rd), \
+      orc_arm_reg_name (Rm), orc_arm_reg_name(Rs)); \
+  orc_arm_emit (p, orc_arm_mul (cond,S,Rd,Rm,Rs)); \
+} while (0)
 
-#define orc_arm_mull(op,cond,S,RdH,RdL,Rm,Rs) (op|((cond)<<28)|((S)<<20)|((RdH)<<16)|((RdL)<<12)|((Rs)<<8)|(Rm))
-#define orc_arm_emit_smull(p,cond,S,RdL,RdH,Rm,Rs) orc_arm_emit(p,orc_arm_mull (0x00c00090,cond,S,RdH,RdL,Rm,Rs))
-#define orc_arm_emit_umull(p,cond,S,RdL,RdH,Rm,Rs) orc_arm_emit(p,orc_arm_mull (0x00800090,cond,S,RdH,RdL,Rm,Rs))
+#define orc_arm_mull(op,cond,S,RdL,RdH,Rn,Rm) (op|((cond)<<28)|((S)<<20)|(((Rn)&15)<<16)|(((RdL)&15)<<12)|(((RdH)&15)<<8)|((Rm)&15))
+#define orc_arm_emit_smull(p,cond,S,RdL,RdH,Rn,Rm) do { \
+  ORC_ASM_CODE (p, "smull %s, %s, %s, %s\n", orc_arm_reg_name (RdL), \
+      orc_arm_reg_name (RdH), \
+      orc_arm_reg_name (Rn), orc_arm_reg_name(Rm)); \
+  orc_arm_emit(p,orc_arm_mull (0x00c00090,cond,S,RdL,Rm,RdH,Rn)); \
+} while (0)
+#define orc_arm_emit_umull(p,cond,S,RdL,RdH,Rn,Rm) do { \
+  ORC_ASM_CODE (p, "umull %s, %s, %s, %s\n", orc_arm_reg_name (RdL), \
+      orc_arm_reg_name (RdH), \
+      orc_arm_reg_name (Rn), orc_arm_reg_name(Rm)); \
+  orc_arm_emit(p,orc_arm_mull (0x00800090,cond,S,RdL,RdH,Rn,Rm)); \
+} while (0)
+
+void
+orc_arm_loadb (OrcCompiler *compiler, int dest, int src1, int offset)
+{
+  orc_uint32 code;
+
+  code = 0xe5d00000;
+  code |= (src1&0xf) << 16;
+  code |= (dest&0xf) << 12;
+  code |= (offset&0xf0) << 4;
+  code |= offset&0x0f;
+
+  ORC_ASM_CODE(compiler,"  ldrb %s, [%s, #%d]\n",
+      orc_arm_reg_name (dest),
+      orc_arm_reg_name (src1), offset);
+  orc_arm_emit (compiler, code);
+}
+
+void
+orc_arm_storeb (OrcCompiler *compiler, int dest, int offset, int src1)
+{
+  orc_uint32 code;
+
+  code = 0xe5c00000;
+  code |= (dest&0xf) << 16;
+  code |= (src1&0xf) << 12;
+  code |= (offset&0xf0) << 4;
+  code |= offset&0x0f;
+
+  ORC_ASM_CODE(compiler,"  strb %s, [%s, #%d]\n",
+      orc_arm_reg_name (src1),
+      orc_arm_reg_name (dest), offset);
+  orc_arm_emit (compiler, code);
+}
+
+void
+orc_arm_loadl (OrcCompiler *compiler, int dest, int src1, int offset)
+{
+  orc_uint32 code;
+
+  code = 0xe5900000;
+  code |= (src1&0xf) << 16;
+  code |= (dest&0xf) << 12;
+  code |= (offset&0xf0) << 4;
+  code |= offset&0x0f;
+
+  ORC_ASM_CODE(compiler,"  ldr %s, [%s, #%d]\n",
+      orc_arm_reg_name (dest),
+      orc_arm_reg_name (src1), offset);
+  orc_arm_emit (compiler, code);
+}
+
+void
+orc_arm_storel (OrcCompiler *compiler, int dest, int offset, int src1)
+{
+  orc_uint32 code;
+
+  code = 0xe5800000;
+  code |= (dest&0xf) << 16;
+  code |= (src1&0xf) << 12;
+  code |= (offset&0xf0) << 4;
+  code |= offset&0x0f;
+
+  ORC_ASM_CODE(compiler,"  str %s, [%s, #%d]\n",
+      orc_arm_reg_name (src1),
+      orc_arm_reg_name (dest), offset);
+  orc_arm_emit (compiler, code);
+}
 
 void
 orc_arm_loadw (OrcCompiler *compiler, int dest, int src1, int offset)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   code = 0xe1d000b0;
   code |= (src1&0xf) << 16;
@@ -68,7 +166,7 @@ orc_arm_loadw (OrcCompiler *compiler, int dest, int src1, int offset)
 void
 orc_arm_storew (OrcCompiler *compiler, int dest, int offset, int src1)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   code = 0xe1c000b0;
   code |= (dest&0xf) << 16;
@@ -134,14 +232,24 @@ BINARY_MM (addb, sadd8);
 BINARY_MM (addssb, qadd8);
 BINARY_MM (addusb, uqadd8);
 BINARY_DP (andX, and);
-BINARY_DP (andnX, bic);
+
+static void
+arm_rule_andnX (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  int src1 = ORC_SRC_ARG (p, insn, 0);
+  int src2 = ORC_SRC_ARG (p, insn, 1);
+  int dest = ORC_DEST_ARG (p, insn, 0);
+
+  orc_arm_emit_bic_r (p, ORC_ARM_COND_AL, 0, dest, src2, src1); 
+}
+
 static void
 arm_rule_avgX (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
   int src1 = ORC_SRC_ARG (p, insn, 0);
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int mask = p->tmpreg;
-  int tmp = ORC_ARM_IP;
+  int tmp = p->gp_tmpreg;
   int dest = ORC_DEST_ARG (p, insn, 0);
   int type = ORC_PTR_TO_INT(user);
 
@@ -301,7 +409,7 @@ arm_rule_mullb (OrcCompiler *p, void *user, OrcInstruction *insn)
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int tmp1 = p->tmpreg;
-  int tmp2 = ORC_ARM_IP;
+  int tmp2 = p->gp_tmpreg;
   int tmp3 = ORC_ARM_V8;
   int loop = 1;
 
@@ -347,7 +455,7 @@ arm_rule_mulhsb (OrcCompiler *p, void *user, OrcInstruction *insn)
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int tmp1 = p->tmpreg;
-  int tmp2 = ORC_ARM_IP;
+  int tmp2 = p->gp_tmpreg;
   int tmp3 = ORC_ARM_V8;
   int loop = 1;
 
@@ -391,12 +499,12 @@ arm_rule_mulhsb (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 arm_rule_mulhub (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  /* BINARY_UB(mulhub, "((uint32_t)(uint8_t)%s * (uint32_t)(uint8_t)%s) >> 8") */
+  /* BINARY_UB(mulhub, "((orc_uint32)(uint8_t)%s * (orc_uint32)(uint8_t)%s) >> 8") */
   int src1 = ORC_SRC_ARG (p, insn, 0);
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int tmp1 = p->tmpreg;
-  int tmp2 = ORC_ARM_IP;
+  int tmp2 = p->gp_tmpreg;
   int tmp3 = ORC_ARM_V8;
   int loop = 1;
 
@@ -517,7 +625,7 @@ arm_rule_shrsX (OrcCompiler *p, void *user, OrcInstruction *insn)
   int src1 = ORC_SRC_ARG (p, insn, 0);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int mask = p->tmpreg;
-  int tmp = ORC_ARM_IP;
+  int tmp = p->gp_tmpreg;
   int src2type = ORC_SRC_TYPE (p, insn, 1);
   int size = ORC_PTR_TO_INT(user);
   int loop = 4 / size; /* number of items in one register */
@@ -674,7 +782,7 @@ arm_rule_signX (OrcCompiler *p, void *user, OrcInstruction *insn)
   int src1 = ORC_SRC_ARG (p, insn, 0);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int zero = p->tmpreg;
-  int ones = ORC_ARM_IP;
+  int ones = p->gp_tmpreg;
   int tmp = ORC_ARM_V8;
   int type = ORC_PTR_TO_INT(user);
 
@@ -792,12 +900,12 @@ arm_rule_mulhsw (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 arm_rule_mulhuw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  /* BINARY_UW(mulhuw, "((uint32_t)((uint16_t)%s) * (uint32_t)((uint16_t)%s)) >> 16") */
+  /* BINARY_UW(mulhuw, "((orc_uint32)((uint16_t)%s) * (orc_uint32)((uint16_t)%s)) >> 16") */
   int src1 = ORC_SRC_ARG (p, insn, 0);
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int tmp1 = p->tmpreg;
-  int tmp2 = ORC_ARM_V8;
+  int tmp2 = p->gp_tmpreg;
   int loop = 1;
 
   /* extract first halves */
@@ -1206,7 +1314,7 @@ arm_rule_mulsbw (OrcCompiler *p, void *user, OrcInstruction *insn)
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int tmp1 = p->tmpreg;
-  int tmp2 = ORC_ARM_V8;
+  int tmp2 = p->gp_tmpreg;
   int loop = 1;
 
   /* first item */
@@ -1233,7 +1341,7 @@ arm_rule_mulubw (OrcCompiler *p, void *user, OrcInstruction *insn)
   int src2 = ORC_SRC_ARG (p, insn, 1);
   int dest = ORC_DEST_ARG (p, insn, 0);
   int tmp1 = p->tmpreg;
-  int tmp2 = ORC_ARM_V8;
+  int tmp2 = p->gp_tmpreg;
   int loop = 1;
 
   /* first item */
@@ -1351,13 +1459,13 @@ arm_rule_select1wb (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 arm_rule_select0lw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  /* UNARY_LW(select0lw, "(uint32_t)%s & 0xffff") */
+  /* UNARY_LW(select0lw, "(orc_uint32)%s & 0xffff") */
   /* NOP */
 }
 static void
 arm_rule_select1lw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  /* UNARY_LW(select1lw, "((uint32_t)%s >> 16)&0xffff") */
+  /* UNARY_LW(select1lw, "((orc_uint32)%s >> 16)&0xffff") */
   int src1 = ORC_SRC_ARG (p, insn, 0);
   int dest = ORC_DEST_ARG (p, insn, 0);
 
@@ -1382,6 +1490,8 @@ arm_rule_swapl (OrcCompiler *p, void *user, OrcInstruction *insn)
   orc_arm_emit_rev (p, ORC_ARM_COND_AL, dest, src1);
 }
 
+#define FAIL if (0)
+
 void
 orc_compiler_orc_arm_register_rules (OrcTarget *target)
 {
@@ -1389,7 +1499,7 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
 
   rule_set = orc_rule_set_new (orc_opcode_set_get("sys"), target, 0);
 
-  orc_rule_register (rule_set, "absb", arm_rule_absX, (void *)0);
+  FAIL orc_rule_register (rule_set, "absb", arm_rule_absX, (void *)0);
   orc_rule_register (rule_set, "addb", arm_rule_addb, NULL);
   orc_rule_register (rule_set, "addssb", arm_rule_addssb, NULL);
   orc_rule_register (rule_set, "addusb", arm_rule_addusb, NULL);
@@ -1400,59 +1510,59 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
   orc_rule_register (rule_set, "cmpeqb", arm_rule_cmpeqX, (void *)1);
   orc_rule_register (rule_set, "cmpgtsb", arm_rule_cmpgtsX, (void *)1);
   orc_rule_register (rule_set, "copyb", arm_rule_copyX, NULL);
-  orc_rule_register (rule_set, "maxsb", arm_rule_maxsb, NULL);
-  orc_rule_register (rule_set, "maxub", arm_rule_maxub, NULL);
-  orc_rule_register (rule_set, "minsb", arm_rule_minsb, NULL);
-  orc_rule_register (rule_set, "minub", arm_rule_minub, NULL);
+  FAIL orc_rule_register (rule_set, "maxsb", arm_rule_maxsb, NULL);
+  FAIL orc_rule_register (rule_set, "maxub", arm_rule_maxub, NULL);
+  FAIL orc_rule_register (rule_set, "minsb", arm_rule_minsb, NULL);
+  FAIL orc_rule_register (rule_set, "minub", arm_rule_minub, NULL);
   orc_rule_register (rule_set, "orb", arm_rule_orX, NULL);
   orc_rule_register (rule_set, "shlb", arm_rule_shlX, (void *)1);
-  orc_rule_register (rule_set, "shrsb", arm_rule_shrsX, (void *)1);
-  orc_rule_register (rule_set, "shrub", arm_rule_shruX, (void *)1);
-  orc_rule_register (rule_set, "signb", arm_rule_signX, (void *)0);
+  FAIL orc_rule_register (rule_set, "shrsb", arm_rule_shrsX, (void *)1);
+  FAIL orc_rule_register (rule_set, "shrub", arm_rule_shruX, (void *)1);
+  FAIL orc_rule_register (rule_set, "signb", arm_rule_signX, (void *)0);
   orc_rule_register (rule_set, "subb", arm_rule_subb, NULL);
   orc_rule_register (rule_set, "subssb", arm_rule_subssb, NULL);
   orc_rule_register (rule_set, "subusb", arm_rule_subusb, NULL);
   orc_rule_register (rule_set, "xorb", arm_rule_xorX, NULL);
   orc_rule_register (rule_set, "mullb", arm_rule_mullb, NULL);
   orc_rule_register (rule_set, "mulhsb", arm_rule_mulhsb, NULL);
-  orc_rule_register (rule_set, "mulhub", arm_rule_mulhub, NULL);
+  FAIL orc_rule_register (rule_set, "mulhub", arm_rule_mulhub, NULL);
 
-  orc_rule_register (rule_set, "absw", arm_rule_absX, (void *)1);
+  FAIL orc_rule_register (rule_set, "absw", arm_rule_absX, (void *)1);
   orc_rule_register (rule_set, "addw", arm_rule_addw, NULL);
   orc_rule_register (rule_set, "addssw", arm_rule_addssw, NULL);
   orc_rule_register (rule_set, "addusw", arm_rule_addusw, NULL);
   orc_rule_register (rule_set, "andw", arm_rule_andX, NULL);
   orc_rule_register (rule_set, "andnw", arm_rule_andnX, NULL);
-  orc_rule_register (rule_set, "avgsw", arm_rule_avgX, (void *)2);
+  FAIL orc_rule_register (rule_set, "avgsw", arm_rule_avgX, (void *)2);
   orc_rule_register (rule_set, "avguw", arm_rule_avgX, (void *)1);
   orc_rule_register (rule_set, "cmpeqw", arm_rule_cmpeqX, (void *)2);
   orc_rule_register (rule_set, "cmpgtsw", arm_rule_cmpgtsX, (void *)2);
   orc_rule_register (rule_set, "copyw", arm_rule_copyX, NULL);
-  orc_rule_register (rule_set, "maxsw", arm_rule_maxsw, NULL);
-  orc_rule_register (rule_set, "maxuw", arm_rule_maxuw, NULL);
-  orc_rule_register (rule_set, "minsw", arm_rule_minsw, NULL);
-  orc_rule_register (rule_set, "minuw", arm_rule_minuw, NULL);
+  FAIL orc_rule_register (rule_set, "maxsw", arm_rule_maxsw, NULL);
+  FAIL orc_rule_register (rule_set, "maxuw", arm_rule_maxuw, NULL);
+  FAIL orc_rule_register (rule_set, "minsw", arm_rule_minsw, NULL);
+  FAIL orc_rule_register (rule_set, "minuw", arm_rule_minuw, NULL);
   orc_rule_register (rule_set, "orw", arm_rule_orX, NULL);
   orc_rule_register (rule_set, "shlw", arm_rule_shlX, (void *)2);
-  orc_rule_register (rule_set, "shrsw", arm_rule_shrsX, (void *)2);
+  FAIL orc_rule_register (rule_set, "shrsw", arm_rule_shrsX, (void *)2);
   orc_rule_register (rule_set, "shruw", arm_rule_shruX, (void *)2);
-  orc_rule_register (rule_set, "signw", arm_rule_signX, (void *)1);
+  FAIL orc_rule_register (rule_set, "signw", arm_rule_signX, (void *)1);
   orc_rule_register (rule_set, "subw", arm_rule_subw, NULL);
   orc_rule_register (rule_set, "subssw", arm_rule_subssw, NULL);
   orc_rule_register (rule_set, "subusw", arm_rule_subusw, NULL);
   orc_rule_register (rule_set, "xorw", arm_rule_xorX, NULL);
   orc_rule_register (rule_set, "mullw", arm_rule_mullw, NULL);
   orc_rule_register (rule_set, "mulhsw", arm_rule_mulhsw, NULL);
-  orc_rule_register (rule_set, "mulhuw", arm_rule_mulhuw, NULL);
+  FAIL orc_rule_register (rule_set, "mulhuw", arm_rule_mulhuw, NULL);
 
-  orc_rule_register (rule_set, "absl", arm_rule_absl, NULL);
+  FAIL orc_rule_register (rule_set, "absl", arm_rule_absl, NULL);
   orc_rule_register (rule_set, "addl", arm_rule_addl, NULL);
   orc_rule_register (rule_set, "addssl", arm_rule_addssl, NULL);
   orc_rule_register (rule_set, "addusl", arm_rule_addusl, NULL);
   orc_rule_register (rule_set, "andl", arm_rule_andX, NULL);
   orc_rule_register (rule_set, "andnl", arm_rule_andnX, NULL);
-  orc_rule_register (rule_set, "avgul", arm_rule_avgXl, NULL);
-  orc_rule_register (rule_set, "avgsl", arm_rule_avgXl, NULL);
+  FAIL orc_rule_register (rule_set, "avgul", arm_rule_avgXl, NULL);
+  FAIL orc_rule_register (rule_set, "avgsl", arm_rule_avgXl, NULL);
   orc_rule_register (rule_set, "cmpeql", arm_rule_cmpeql, NULL);
   orc_rule_register (rule_set, "cmpgtsl", arm_rule_cmpgtsl, NULL);
   orc_rule_register (rule_set, "copyl", arm_rule_copyX, NULL);
@@ -1462,39 +1572,39 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
   orc_rule_register (rule_set, "minul", arm_rule_minul, NULL);
   orc_rule_register (rule_set, "mulll", arm_rule_mulll, NULL);
   orc_rule_register (rule_set, "mulhsl", arm_rule_mulhsl, NULL);
-  orc_rule_register (rule_set, "mulhul", arm_rule_mulhul, NULL);
+  FAIL orc_rule_register (rule_set, "mulhul", arm_rule_mulhul, NULL);
   orc_rule_register (rule_set, "orl", arm_rule_orX, NULL);
   orc_rule_register (rule_set, "shll", arm_rule_shlX, (void *)4);
   orc_rule_register (rule_set, "shrsl", arm_rule_shrsX, (void *)4);
   orc_rule_register (rule_set, "shrul", arm_rule_shruX, (void *)4);
-  orc_rule_register (rule_set, "signl", arm_rule_signl, NULL);
+  FAIL orc_rule_register (rule_set, "signl", arm_rule_signl, NULL);
   orc_rule_register (rule_set, "subl", arm_rule_subl, NULL);
-  orc_rule_register (rule_set, "subssl", arm_rule_subssl, NULL);
+  FAIL orc_rule_register (rule_set, "subssl", arm_rule_subssl, NULL);
   orc_rule_register (rule_set, "subusl", arm_rule_subusl, NULL);
   orc_rule_register (rule_set, "xorl", arm_rule_xorX, NULL);
 
   orc_rule_register (rule_set, "convsbw", arm_rule_convsbw, NULL);
-  orc_rule_register (rule_set, "convubw", arm_rule_convubw, NULL);
-  orc_rule_register (rule_set, "convswl", arm_rule_convswl, NULL);
-  orc_rule_register (rule_set, "convuwl", arm_rule_convuwl, NULL);
+  FAIL orc_rule_register (rule_set, "convubw", arm_rule_convubw, NULL);
+  FAIL orc_rule_register (rule_set, "convswl", arm_rule_convswl, NULL);
+  FAIL orc_rule_register (rule_set, "convuwl", arm_rule_convuwl, NULL);
   orc_rule_register (rule_set, "convwb", arm_rule_convwb, NULL);
   orc_rule_register (rule_set, "convssswb", arm_rule_convssswb, NULL);
   orc_rule_register (rule_set, "convsuswb", arm_rule_convsuswb, NULL);
-  orc_rule_register (rule_set, "convusswb", arm_rule_convusswb, NULL);
-  orc_rule_register (rule_set, "convuuswb", arm_rule_convuuswb, NULL);
+  FAIL orc_rule_register (rule_set, "convusswb", arm_rule_convusswb, NULL);
+  FAIL orc_rule_register (rule_set, "convuuswb", arm_rule_convuuswb, NULL);
   orc_rule_register (rule_set, "convlw", arm_rule_convlw, NULL);
-  orc_rule_register (rule_set, "convssslw", arm_rule_convssslw, NULL);
+  FAIL orc_rule_register (rule_set, "convssslw", arm_rule_convssslw, NULL);
   orc_rule_register (rule_set, "convsuslw", arm_rule_convsuslw, NULL);
-  orc_rule_register (rule_set, "convusslw", arm_rule_convusslw, NULL);
-  orc_rule_register (rule_set, "convuuslw", arm_rule_convuuslw, NULL);
+  FAIL orc_rule_register (rule_set, "convusslw", arm_rule_convusslw, NULL);
+  FAIL orc_rule_register (rule_set, "convuuslw", arm_rule_convuuslw, NULL);
 
   orc_rule_register (rule_set, "mulsbw", arm_rule_mulsbw, NULL);
-  orc_rule_register (rule_set, "mulubw", arm_rule_mulubw, NULL);
-  orc_rule_register (rule_set, "mulswl", arm_rule_mulswl, NULL);
-  orc_rule_register (rule_set, "muluwl", arm_rule_muluwl, NULL);
+  FAIL orc_rule_register (rule_set, "mulubw", arm_rule_mulubw, NULL);
+  FAIL orc_rule_register (rule_set, "mulswl", arm_rule_mulswl, NULL);
+  FAIL orc_rule_register (rule_set, "muluwl", arm_rule_muluwl, NULL);
 
   orc_rule_register (rule_set, "mergewl", arm_rule_mergewl, NULL);
-  orc_rule_register (rule_set, "mergebw", arm_rule_mergebw, NULL);
+  FAIL orc_rule_register (rule_set, "mergebw", arm_rule_mergebw, NULL);
   orc_rule_register (rule_set, "select0wb", arm_rule_select0wb, NULL);
   orc_rule_register (rule_set, "select1wb", arm_rule_select1wb, NULL);
   orc_rule_register (rule_set, "select0lw", arm_rule_select0lw, NULL);
