@@ -5,7 +5,6 @@
 #include <string.h>
 #include <stdlib.h>
 
-#include <unistd.h>
 #include <sys/types.h>
 
 #include <orc/orcprogram.h>
@@ -206,10 +205,41 @@ orc_neon_emit_mov (OrcCompiler *compiler, int dest, int src)
       dest, src, src);
 }
 
+void
+orc_neon_emit_mov_quad (OrcCompiler *compiler, int dest, int src)
+{
+  orc_neon_emit_binary_quad (compiler, "vorr", 0xf2200110,
+      dest, src, src);
+}
+
+void
+orc_neon_preload (OrcCompiler *compiler, OrcVariable *var, int write,
+    int offset)
+{
+  orc_uint32 code;
+
+  /* Don't use multiprocessing extensions */
+  write = FALSE;
+
+  ORC_ASM_CODE(compiler,"  pld%s [%s, #%d]\n",
+      write ? "w" : "",
+      orc_arm_reg_name (var->ptr_register), offset);
+  code = 0xf510f000;
+  if (!write) code |= (1<<22);
+  code |= (var->ptr_register&0xf) << 16;
+  if (offset < 0) {
+    code |= ((-offset)&0xfff) << 0;
+  } else {
+    code |= (offset&0xfff) << 0;
+    code |= (1<<23);
+  }
+  orc_arm_emit (compiler, code);
+}
+
 void
 orc_neon_load_halfvec_aligned (OrcCompiler *compiler, OrcVariable *var, int update)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   ORC_ASM_CODE(compiler,"  vld1.32 %s[0], [%s]%s\n",
       orc_neon_reg_name (var->alloc),
@@ -226,7 +256,7 @@ orc_neon_load_halfvec_aligned (OrcCompiler *compiler, OrcVariable *var, int upda
 void
 orc_neon_load_vec_aligned (OrcCompiler *compiler, OrcVariable *var, int update)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   ORC_ASM_CODE(compiler,"  vld1.64 %s, [%s]%s\n",
       orc_neon_reg_name (var->alloc),
@@ -244,8 +274,20 @@ void
 orc_neon_load_vec_unaligned (OrcCompiler *compiler, OrcVariable *var,
     int update)
 {
-  uint32_t code;
+  orc_uint32 code;
 
+  ORC_ASM_CODE(compiler,"  vld1.8 %s, [%s]%s\n",
+      orc_neon_reg_name (var->alloc),
+      orc_arm_reg_name (var->ptr_register),
+      update ? "!" : "");
+  code = 0xf420070d;
+  code |= (var->ptr_register&0xf) << 16;
+  code |= ((var->alloc)&0xf) << 12;
+  code |= (((var->alloc)>>4)&0x1) << 22;
+  code |= (!update) << 1;
+  orc_arm_emit (compiler, code);
+#if 0
+  /* used with need_mask_regs */
   ORC_ASM_CODE(compiler,"  vld1.64 %s, [%s]%s\n",
       orc_neon_reg_name (var->aligned_data + 1),
       orc_arm_reg_name (var->ptr_register),
@@ -268,14 +310,32 @@ orc_neon_load_vec_unaligned (OrcCompiler *compiler, OrcVariable *var,
 //orc_neon_emit_mov (compiler, var->alloc, var->mask_alloc);
 
   orc_neon_emit_mov (compiler, var->aligned_data, var->aligned_data + 1);
+#endif
 }
 
 void
 orc_neon_load_halfvec_unaligned (OrcCompiler *compiler, OrcVariable *var,
     int update)
 {
-  uint32_t code;
+  orc_uint32 code;
 
+  ORC_ASM_CODE(compiler,"  vld1.8 %s, [%s]\n",
+      orc_neon_reg_name (var->alloc),
+      orc_arm_reg_name (var->ptr_register));
+  code = 0xf420070d;
+  code |= (var->ptr_register&0xf) << 16;
+  code |= ((var->alloc)&0xf) << 12;
+  code |= (((var->alloc)>>4)&0x1) << 22;
+  //code |= (!update) << 1;
+  code |= (1) << 1;
+  orc_arm_emit (compiler, code);
+
+  if (update) {
+    orc_arm_emit_add_imm (compiler, var->ptr_register,
+        var->ptr_register, 4);
+  }
+#if 0
+  /* used with need_mask_regs */
   ORC_ASM_CODE(compiler,"  vld1.32 %s[1], [%s]%s\n",
       orc_neon_reg_name (var->aligned_data),
       orc_arm_reg_name (var->ptr_register),
@@ -297,18 +357,105 @@ orc_neon_load_halfvec_unaligned (OrcCompiler *compiler, OrcVariable *var,
 
   orc_neon_emit_unary (compiler, "vrev64.i32", 0xf3b80000,
       var->aligned_data, var->aligned_data);
+#endif
+}
+
+void
+orc_neon_load_fourvec_aligned (OrcCompiler *compiler, OrcVariable *var, int update)
+{
+  orc_uint32 code;
+
+  ORC_ASM_CODE(compiler,"  vld1.64 { %s, %s, %s, %s }, [%s,:256]%s\n",
+      orc_neon_reg_name (var->alloc),
+      orc_neon_reg_name (var->alloc + 1),
+      orc_neon_reg_name (var->alloc + 2),
+      orc_neon_reg_name (var->alloc + 3),
+      orc_arm_reg_name (var->ptr_register),
+      update ? "!" : "");
+  code = 0xf42002dd;
+  code |= (var->ptr_register&0xf) << 16;
+  code |= (var->alloc&0xf) << 12;
+  code |= ((var->alloc>>4)&0x1) << 22;
+  code |= (!update) << 1;
+  orc_arm_emit (compiler, code);
+}
+
+void
+orc_neon_load_fourvec_unaligned (OrcCompiler *compiler, OrcVariable *var,
+    int update)
+{
+  orc_uint32 code;
+
+  ORC_ASM_CODE(compiler,"  vld1.8 { %s, %s, %s, %s }, [%s]%s\n",
+      orc_neon_reg_name (var->alloc),
+      orc_neon_reg_name (var->alloc + 1),
+      orc_neon_reg_name (var->alloc + 2),
+      orc_neon_reg_name (var->alloc + 3),
+      orc_arm_reg_name (var->ptr_register),
+      update ? "!" : "");
+  code = 0xf420020d;
+  code |= (var->ptr_register&0xf) << 16;
+  code |= ((var->alloc)&0xf) << 12;
+  code |= (((var->alloc)>>4)&0x1) << 22;
+  code |= (!update) << 1;
+  orc_arm_emit (compiler, code);
+}
+
+void
+orc_neon_load_twovec_aligned (OrcCompiler *compiler, OrcVariable *var, int update)
+{
+  orc_uint32 code;
+
+  ORC_ASM_CODE(compiler,"  vld1.64 { %s, %s }, [%s,:128]%s\n",
+      orc_neon_reg_name (var->alloc),
+      orc_neon_reg_name (var->alloc + 1),
+      orc_arm_reg_name (var->ptr_register),
+      update ? "!" : "");
+  code = 0xf4200aed;
+  code |= (var->ptr_register&0xf) << 16;
+  code |= (var->alloc&0xf) << 12;
+  code |= ((var->alloc>>4)&0x1) << 22;
+  code |= (!update) << 1;
+  orc_arm_emit (compiler, code);
+}
+
+void
+orc_neon_load_twovec_unaligned (OrcCompiler *compiler, OrcVariable *var,
+    int update)
+{
+  orc_uint32 code;
+
+  ORC_ASM_CODE(compiler,"  vld1.8 { %s, %s }, [%s]%s\n",
+      orc_neon_reg_name (var->alloc),
+      orc_neon_reg_name (var->alloc + 1),
+      orc_arm_reg_name (var->ptr_register),
+      update ? "!" : "");
+  code = 0xf4200a0d;
+  code |= (var->ptr_register&0xf) << 16;
+  code |= ((var->alloc)&0xf) << 12;
+  code |= (((var->alloc)>>4)&0x1) << 22;
+  code |= (!update) << 1;
+  orc_arm_emit (compiler, code);
 }
 
 void
 orc_neon_loadb (OrcCompiler *compiler, OrcVariable *var, int update)
 {
-  uint32_t code;
+  orc_uint32 code;
   int i;
 
-  if (var->is_aligned && compiler->loop_shift == 3) {
+  if (var->is_aligned && compiler->loop_shift == 5) {
+    orc_neon_load_fourvec_aligned (compiler, var, update);
+  } else if (var->is_aligned && compiler->loop_shift == 4) {
+    orc_neon_load_twovec_aligned (compiler, var, update);
+  } else if (var->is_aligned && compiler->loop_shift == 3) {
     orc_neon_load_vec_aligned (compiler, var, update);
   } else if (var->is_aligned && compiler->loop_shift == 2) {
     orc_neon_load_halfvec_aligned (compiler, var, update);
+  } else if (compiler->loop_shift == 5) {
+    orc_neon_load_fourvec_unaligned (compiler, var, update);
+  } else if (compiler->loop_shift == 4) {
+    orc_neon_load_twovec_unaligned (compiler, var, update);
   } else if (compiler->loop_shift == 3) {
     orc_neon_load_vec_unaligned (compiler, var, update);
   } else if (compiler->loop_shift == 2) {
@@ -334,14 +481,20 @@ orc_neon_loadb (OrcCompiler *compiler, OrcVariable *var, int update)
 void
 orc_neon_loadw (OrcCompiler *compiler, OrcVariable *var, int update)
 {
-  if (var->is_aligned && compiler->loop_shift == 2) {
+  if (var->is_aligned && compiler->loop_shift == 3) {
+    orc_neon_load_twovec_aligned (compiler, var, update);
+  } else if (var->is_aligned && compiler->loop_shift == 2) {
     orc_neon_load_vec_aligned (compiler, var, update);
   } else if (var->is_aligned && compiler->loop_shift == 1) {
     orc_neon_load_halfvec_aligned (compiler, var, update);
-  } else if (compiler->loop_shift == 2 && var->mask_alloc) {
+  } else if (compiler->loop_shift == 3) {
+    orc_neon_load_twovec_unaligned (compiler, var, update);
+  } else if (compiler->loop_shift == 2) {
     orc_neon_load_vec_unaligned (compiler, var, update);
+  } else if (compiler->loop_shift == 1) {
+    orc_neon_load_halfvec_unaligned (compiler, var, update);
   } else {
-    uint32_t code;
+    orc_uint32 code;
     int i;
 
     if (compiler->loop_shift == 2) {
@@ -371,11 +524,17 @@ orc_neon_loadw (OrcCompiler *compiler, OrcVariable *var, int update)
 void
 orc_neon_loadl (OrcCompiler *compiler, OrcVariable *var, int update)
 {
-  uint32_t code;
+  orc_uint32 code;
   int i;
 
-  if (var->is_aligned && compiler->loop_shift == 1) {
+  if (var->is_aligned && compiler->loop_shift == 2) {
+    orc_neon_load_twovec_aligned (compiler, var, update);
+  } else if (var->is_aligned && compiler->loop_shift == 1) {
     orc_neon_load_vec_aligned (compiler, var, update);
+  } else if (compiler->loop_shift == 2) {
+    orc_neon_load_twovec_unaligned (compiler, var, update);
+  } else if (compiler->loop_shift == 1) {
+    orc_neon_load_vec_unaligned (compiler, var, update);
   } else {
     if (compiler->loop_shift > 0) {
       //ORC_ERROR("slow load");
@@ -397,7 +556,7 @@ orc_neon_loadl (OrcCompiler *compiler, OrcVariable *var, int update)
 void
 orc_neon_loadq (OrcCompiler *compiler, int dest, int src1, int update, int is_aligned)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   ORC_ASM_CODE(compiler,"  vld1.64 %s, [%s]%s\n",
       orc_neon_reg_name (dest),
@@ -415,15 +574,67 @@ orc_neon_loadq (OrcCompiler *compiler, int dest, int src1, int update, int is_al
 void
 orc_neon_storeb (OrcCompiler *compiler, int dest, int update, int src1, int is_aligned)
 {
-  uint32_t code;
+  orc_uint32 code;
   int i;
 
-  if (is_aligned && compiler->loop_shift == 3) {
-    ORC_ASM_CODE(compiler,"  vst1.64 %s, [%s]%s\n",
+  if (is_aligned && compiler->loop_shift == 5) {
+    ORC_ASM_CODE(compiler,"  vst1.8 { %s, %s, %s, %s }, [%s,:256]%s\n",
+        orc_neon_reg_name (src1),
+        orc_neon_reg_name (src1+1),
+        orc_neon_reg_name (src1+2),
+        orc_neon_reg_name (src1+3),
+        orc_arm_reg_name (dest),
+        update ? "!" : "");
+    code = 0xf400023d;
+    code |= (dest&0xf) << 16;
+    code |= (src1&0xf) << 12;
+    code |= ((src1>>4)&0x1) << 22;
+    code |= (!update) << 1;
+    orc_arm_emit (compiler, code);
+  } else if (compiler->loop_shift == 5) {
+    ORC_ASM_CODE(compiler,"  vst1.8 { %s, %s, %s, %s }, [%s]%s\n",
+        orc_neon_reg_name (src1),
+        orc_neon_reg_name (src1+1),
+        orc_neon_reg_name (src1+2),
+        orc_neon_reg_name (src1+3),
+        orc_arm_reg_name (dest),
+        update ? "!" : "");
+    code = 0xf400020d;
+    code |= (dest&0xf) << 16;
+    code |= (src1&0xf) << 12;
+    code |= ((src1>>4)&0x1) << 22;
+    code |= (!update) << 1;
+    orc_arm_emit (compiler, code);
+  } else if (is_aligned && compiler->loop_shift == 4) {
+    ORC_ASM_CODE(compiler,"  vst1.8 { %s, %s }, [%s,:128]%s\n",
+        orc_neon_reg_name (src1),
+        orc_neon_reg_name (src1+1),
+        orc_arm_reg_name (dest),
+        update ? "!" : "");
+    code = 0xf4000a2d;
+    code |= (dest&0xf) << 16;
+    code |= (src1&0xf) << 12;
+    code |= ((src1>>4)&0x1) << 22;
+    code |= (!update) << 1;
+    orc_arm_emit (compiler, code);
+  } else if (compiler->loop_shift == 4) {
+    ORC_ASM_CODE(compiler,"  vst1.8 { %s, %s }, [%s]%s\n",
+        orc_neon_reg_name (src1),
+        orc_neon_reg_name (src1+1),
+        orc_arm_reg_name (dest),
+        update ? "!" : "");
+    code = 0xf4000a0d;
+    code |= (dest&0xf) << 16;
+    code |= (src1&0xf) << 12;
+    code |= ((src1>>4)&0x1) << 22;
+    code |= (!update) << 1;
+    orc_arm_emit (compiler, code);
+  } else if (is_aligned && compiler->loop_shift == 3) {
+    ORC_ASM_CODE(compiler,"  vst1.8 %s, [%s,:64]%s\n",
         orc_neon_reg_name (src1),
         orc_arm_reg_name (dest),
         update ? "!" : "");
-    code = 0xf40007cd;
+    code = 0xf400071d;
     code |= (dest&0xf) << 16;
     code |= (src1&0xf) << 12;
     code |= ((src1>>4)&0x1) << 22;
@@ -449,15 +660,27 @@ orc_neon_storeb (OrcCompiler *compiler, int dest, int update, int src1, int is_a
 void
 orc_neon_storew (OrcCompiler *compiler, int dest, int update, int src1, int is_aligned)
 {
-  uint32_t code;
+  orc_uint32 code;
   int i;
 
-  if (is_aligned && compiler->loop_shift == 2) {
-    ORC_ASM_CODE(compiler,"  vst1.64 %s, [%s]%s\n",
+  if (is_aligned && compiler->loop_shift == 3) {
+    ORC_ASM_CODE(compiler,"  vst1.16 { %s, %s }, [%s,:128]%s\n",
+        orc_neon_reg_name (src1),
+        orc_neon_reg_name (src1 + 1),
+        orc_arm_reg_name (dest),
+        update ? "!" : "");
+    code = 0xf4000a6d;
+    code |= (dest&0xf) << 16;
+    code |= (src1&0xf) << 12;
+    code |= ((src1>>4)&0x1) << 22;
+    code |= (!update) << 1;
+    orc_arm_emit (compiler, code);
+  } else if (is_aligned && compiler->loop_shift == 2) {
+    ORC_ASM_CODE(compiler,"  vst1.16 %s, [%s,:64]%s\n",
         orc_neon_reg_name (src1),
         orc_arm_reg_name (dest),
         update ? "!" : "");
-    code = 0xf40007cd;
+    code = 0xf400075d;
     code |= (dest&0xf) << 16;
     code |= (src1&0xf) << 12;
     code |= ((src1>>4)&0x1) << 22;
@@ -483,15 +706,27 @@ orc_neon_storew (OrcCompiler *compiler, int dest, int update, int src1, int is_a
 void
 orc_neon_storel (OrcCompiler *compiler, int dest, int update, int src1, int is_aligned)
 {
-  uint32_t code;
+  orc_uint32 code;
   int i;
 
   if (is_aligned && compiler->loop_shift == 2) {
-    ORC_ASM_CODE(compiler,"  vst1.64 %s, [%s]%s\n",
+    ORC_ASM_CODE(compiler,"  vst1.32 { %s, %s }, [%s,:128]%s\n",
+        orc_neon_reg_name (src1),
+        orc_neon_reg_name (src1 + 1),
+        orc_arm_reg_name (dest),
+        update ? "!" : "");
+    code = 0xf4000aad;
+    code |= (dest&0xf) << 16;
+    code |= (src1&0xf) << 12;
+    code |= ((src1>>4)&0x1) << 22;
+    code |= (!update) << 1;
+    orc_arm_emit (compiler, code);
+  } else if (is_aligned && compiler->loop_shift == 1) {
+    ORC_ASM_CODE(compiler,"  vst1.32 %s, [%s,:64]%s\n",
         orc_neon_reg_name (src1),
         orc_arm_reg_name (dest),
         update ? "!" : "");
-    code = 0xf40007cd;
+    code = 0xf400079d;
     code |= (dest&0xf) << 16;
     code |= (src1&0xf) << 12;
     code |= ((src1>>4)&0x1) << 22;
@@ -517,7 +752,7 @@ orc_neon_storel (OrcCompiler *compiler, int dest, int update, int src1, int is_a
 void
 orc_neon_storeq (OrcCompiler *compiler, int dest, int update, int src1, int is_aligned)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   ORC_ASM_CODE(compiler,"  vst1.64 %s, [%s]%s\n",
       orc_neon_reg_name (src1),
@@ -546,7 +781,7 @@ orc_neon_get_const_shift (unsigned int value)
 void
 orc_neon_emit_loadib (OrcCompiler *compiler, int reg, int value)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   if (value == 0) {
     orc_neon_emit_binary_quad (compiler, "veor", 0xf3000110, reg, reg, reg);
@@ -569,7 +804,7 @@ orc_neon_emit_loadib (OrcCompiler *compiler, int reg, int value)
 void
 orc_neon_emit_loadiw (OrcCompiler *compiler, int reg, int value)
 {
-  uint32_t code;
+  orc_uint32 code;
   int shift;
   int neg = FALSE;
 
@@ -624,7 +859,7 @@ orc_neon_emit_loadiw (OrcCompiler *compiler, int reg, int value)
 void
 orc_neon_emit_loadil (OrcCompiler *compiler, int reg, int value)
 {
-  uint32_t code;
+  orc_uint32 code;
   int shift;
   int neg = FALSE;
 
@@ -679,7 +914,7 @@ orc_neon_emit_loadil (OrcCompiler *compiler, int reg, int value)
 void
 orc_neon_emit_loadpb (OrcCompiler *compiler, int dest, int param)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   orc_arm_emit_add_imm (compiler, compiler->gp_tmpreg,
       compiler->exec_reg, ORC_STRUCT_OFFSET(OrcExecutor, params[param]));
@@ -697,7 +932,7 @@ orc_neon_emit_loadpb (OrcCompiler *compiler, int dest, int param)
 void
 orc_neon_emit_loadpw (OrcCompiler *compiler, int dest, int param)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   orc_arm_emit_add_imm (compiler, compiler->gp_tmpreg,
       compiler->exec_reg, ORC_STRUCT_OFFSET(OrcExecutor, params[param]));
@@ -715,7 +950,7 @@ orc_neon_emit_loadpw (OrcCompiler *compiler, int dest, int param)
 void
 orc_neon_emit_loadpl (OrcCompiler *compiler, int dest, int param)
 {
-  uint32_t code;
+  orc_uint32 code;
 
   orc_arm_emit_add_imm (compiler, compiler->gp_tmpreg,
       compiler->exec_reg, ORC_STRUCT_OFFSET(OrcExecutor, params[param]));
@@ -828,23 +1063,15 @@ orc_neon_rule_ ## opcode (OrcCompiler *p, void *user, OrcInstruction *insn) \
     return; \
   } \
   if (p->loop_shift <= vec_shift) { \
-    uint32_t x = code; \
-    ORC_ASM_CODE(p,"  " insn_name " %s, %s\n", \
-        orc_neon_reg_name (p->vars[insn->dest_args[0]].alloc), \
-        orc_neon_reg_name (p->vars[insn->src_args[0]].alloc)); \
-    x = NEON_BINARY(code, p->vars[insn->dest_args[0]].alloc, \
+    orc_neon_emit_binary (p, "vorr", 0xf2200110, \
+        p->vars[insn->dest_args[0]].alloc, \
         p->vars[insn->src_args[0]].alloc, \
         p->vars[insn->src_args[0]].alloc); \
-    orc_arm_emit (p, x); \
   } else if (p->loop_shift == vec_shift + 1) { \
-    uint32_t x = code; \
-    ORC_ASM_CODE(p,"  " insn_name " %s, %s\n", \
-        orc_neon_reg_name (p->vars[insn->dest_args[0]].alloc), \
-        orc_neon_reg_name (p->vars[insn->src_args[0]].alloc)); \
-    x = NEON_BINARY(code, p->vars[insn->dest_args[0]].alloc, \
+    orc_neon_emit_binary_quad (p, "vorr", 0xf2200110, \
+        p->vars[insn->dest_args[0]].alloc, \
         p->vars[insn->src_args[0]].alloc, \
         p->vars[insn->src_args[0]].alloc); \
-    orc_arm_emit (p, x); \
   } else { \
     ORC_COMPILER_ERROR(p, "shift too large"); \
   } \
@@ -852,7 +1079,7 @@ orc_neon_rule_ ## opcode (OrcCompiler *p, void *user, OrcInstruction *insn) \
 
 
 typedef struct {
-  uint32_t code;
+  orc_uint32 code;
   char *name;
   int negate;
   int bits;
@@ -885,7 +1112,7 @@ static void
 orc_neon_rule_shift (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
   int type = (unsigned long)user;
-  uint32_t code;
+  orc_uint32 code;
 
   if (p->vars[insn->src_args[1]].vartype == ORC_VAR_TYPE_CONST) {
     int shift = p->vars[insn->src_args[1]].value;
@@ -960,7 +1187,7 @@ orc_neon_rule_shift (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 orc_neon_rule_shrsw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  uint32_t code;
+  orc_uint32 code;
   if (p->vars[insn->src_args[1]].vartype == ORC_VAR_TYPE_CONST) {
     code = 0xf2900010;
     ORC_ASM_CODE(p,"  vshr.s16 %s, %s, #%d\n",
@@ -994,7 +1221,7 @@ orc_neon_rule_shrsw (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 orc_neon_rule_shrsl (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  uint32_t code;
+  orc_uint32 code;
   if (p->vars[insn->src_args[1]].vartype == ORC_VAR_TYPE_CONST) {
     code = 0xf2900010;
     ORC_ASM_CODE(p,"  vshr.s32 %s, %s, #%d\n",
@@ -1030,11 +1257,20 @@ orc_neon_rule_shrsl (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 orc_neon_rule_andn (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
+  int max_shift = (long)user;
+
   /* this is special because the operand order is reversed */
-  orc_neon_emit_binary (p, "vbic", 0xf2100110,
-      p->vars[insn->dest_args[0]].alloc,
-      p->vars[insn->src_args[1]].alloc,
-      p->vars[insn->src_args[0]].alloc);
+  if (p->loop_shift <= max_shift) { \
+    orc_neon_emit_binary (p, "vbic", 0xf2100110,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[1]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  } else {
+    orc_neon_emit_binary_quad (p, "vbic", 0xf2100110,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[1]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  }
 }
 
 
@@ -1158,10 +1394,27 @@ UNARY(convlf,"vcvt.f32.s32",0xf3bb0600, 1)
 static void
 orc_neon_rule_accw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  orc_neon_emit_binary (p, "vadd.i16", 0xf2100800,
-      p->vars[insn->dest_args[0]].alloc,
-      p->vars[insn->dest_args[0]].alloc,
-      p->vars[insn->src_args[0]].alloc);
+  unsigned int code;
+
+  if (p->loop_shift < 2) {
+    ORC_ASM_CODE(p,"  vshl.i64 %s, %s, #%d\n",
+        orc_neon_reg_name (p->tmpreg),
+        orc_neon_reg_name (p->vars[insn->src_args[0]].alloc), 48);
+    code = NEON_BINARY(0xf2a00590, p->tmpreg, 0,
+        p->vars[insn->src_args[0]].alloc);
+    code |= (48) << 16;
+    orc_arm_emit (p, code);
+
+    orc_neon_emit_binary (p, "vadd.i16", 0xf2100800,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg);
+  } else {
+    orc_neon_emit_binary (p, "vadd.i16", 0xf2100800,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  }
 }
 
 static void
@@ -1193,9 +1446,15 @@ orc_neon_rule_accl (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 orc_neon_rule_select1wb (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  orc_neon_emit_unary (p, "vrev16.i8", 0xf3b00100,
-      p->vars[insn->dest_args[0]].alloc,
-      p->vars[insn->src_args[0]].alloc);
+  if (p->loop_shift <= 2) {
+    orc_neon_emit_unary (p, "vrev16.i8", 0xf3b00100,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  } else {
+    orc_neon_emit_unary_quad (p, "vrev16.i8", 0xf3b00100,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  }
   orc_neon_emit_unary_narrow (p, "vmovn.i16", 0xf3b20200,
       p->vars[insn->dest_args[0]].alloc,
       p->vars[insn->dest_args[0]].alloc);
@@ -1204,9 +1463,15 @@ orc_neon_rule_select1wb (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 orc_neon_rule_select1lw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  orc_neon_emit_unary (p, "vrev32.i16", 0xf3b40080,
-      p->vars[insn->dest_args[0]].alloc,
-      p->vars[insn->src_args[0]].alloc);
+  if (p->loop_shift <= 1) {
+    orc_neon_emit_unary (p, "vrev32.i16", 0xf3b40080,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  } else {
+    orc_neon_emit_unary_quad (p, "vrev32.i16", 0xf3b40080,
+        p->vars[insn->dest_args[0]].alloc,
+        p->vars[insn->src_args[0]].alloc);
+  }
   orc_neon_emit_unary_narrow (p, "vmovn.i32", 0xf3b60200,
       p->vars[insn->dest_args[0]].alloc,
       p->vars[insn->dest_args[0]].alloc);
@@ -1215,64 +1480,121 @@ orc_neon_rule_select1lw (OrcCompiler *p, void *user, OrcInstruction *insn)
 static void
 orc_neon_rule_mergebw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  if (p->vars[insn->dest_args[0]].alloc != p->vars[insn->src_args[0]].alloc) {
-    orc_neon_emit_mov (p, p->vars[insn->dest_args[0]].alloc,
-        p->vars[insn->src_args[0]].alloc);
-  }
+  if (p->loop_shift <= 2) {
+    if (p->vars[insn->dest_args[0]].alloc != p->vars[insn->src_args[0]].alloc) {
+      orc_neon_emit_mov (p, p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[0]].alloc);
+    }
 
-  if (p->vars[insn->src_args[1]].last_use != p->insn_index) {
-    orc_neon_emit_mov (p, p->tmpreg, p->vars[insn->src_args[1]].alloc);
-    orc_neon_emit_unary (p, "vzip.8", 0xf3b20180,
+    if (p->vars[insn->src_args[1]].last_use != p->insn_index) {
+      orc_neon_emit_mov (p, p->tmpreg, p->vars[insn->src_args[1]].alloc);
+      orc_neon_emit_unary (p, "vzip.8", 0xf3b20180,
+          p->vars[insn->dest_args[0]].alloc,
+          p->tmpreg);
+    } else {
+      orc_neon_emit_unary (p, "vzip.8", 0xf3b20180,
+          p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[1]].alloc);
+    }
+  } else {
+    if (p->vars[insn->dest_args[0]].alloc != p->vars[insn->src_args[0]].alloc) {
+      orc_neon_emit_mov_quad (p, p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[0]].alloc);
+    }
+
+    orc_neon_emit_mov_quad (p, p->tmpreg, p->vars[insn->src_args[1]].alloc);
+    orc_neon_emit_unary_quad (p, "vzip.8", 0xf3b20180,
         p->vars[insn->dest_args[0]].alloc,
         p->tmpreg);
-  } else {
-    orc_neon_emit_unary (p, "vzip.8", 0xf3b20180,
-        p->vars[insn->dest_args[0]].alloc,
-        p->vars[insn->src_args[1]].alloc);
   }
 }
 
 static void
 orc_neon_rule_mergewl (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  if (p->vars[insn->dest_args[0]].alloc != p->vars[insn->src_args[0]].alloc) {
-    orc_neon_emit_mov (p, p->vars[insn->dest_args[0]].alloc,
-        p->vars[insn->src_args[0]].alloc);
-  }
+  if (p->loop_shift <= 1) {
+    if (p->vars[insn->dest_args[0]].alloc != p->vars[insn->src_args[0]].alloc) {
+      orc_neon_emit_mov (p, p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[0]].alloc);
+    }
 
-  if (p->vars[insn->src_args[1]].last_use != p->insn_index) {
-    orc_neon_emit_mov (p, p->tmpreg, p->vars[insn->src_args[1]].alloc);
-    orc_neon_emit_unary (p, "vzip.16", 0xf3b60180,
-        p->vars[insn->dest_args[0]].alloc,
-        p->tmpreg);
+    if (p->vars[insn->src_args[1]].last_use != p->insn_index) {
+      orc_neon_emit_mov (p, p->tmpreg, p->vars[insn->src_args[1]].alloc);
+      orc_neon_emit_unary (p, "vzip.16", 0xf3b60180,
+          p->vars[insn->dest_args[0]].alloc,
+          p->tmpreg);
+    } else {
+      orc_neon_emit_unary (p, "vzip.16", 0xf3b60180,
+          p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[1]].alloc);
+    }
   } else {
-    orc_neon_emit_unary (p, "vzip.16", 0xf3b60180,
-        p->vars[insn->dest_args[0]].alloc,
-        p->vars[insn->src_args[1]].alloc);
+    if (p->vars[insn->dest_args[0]].alloc != p->vars[insn->src_args[0]].alloc) {
+      orc_neon_emit_mov_quad (p, p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[0]].alloc);
+    }
+
+    if (p->vars[insn->src_args[1]].last_use != p->insn_index) {
+      orc_neon_emit_mov_quad (p, p->tmpreg, p->vars[insn->src_args[1]].alloc);
+      orc_neon_emit_unary_quad (p, "vzip.16", 0xf3b60180,
+          p->vars[insn->dest_args[0]].alloc,
+          p->tmpreg);
+    } else {
+      orc_neon_emit_unary_quad (p, "vzip.16", 0xf3b60180,
+          p->vars[insn->dest_args[0]].alloc,
+          p->vars[insn->src_args[1]].alloc);
+    }
   }
 }
 
 static void
 orc_neon_rule_accsadubl (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
-  uint32_t x;
-  
-  x = 0xf3800700;
-  ORC_ASM_CODE(p,"  vabdl.u8 %s, %s, %s\n",
-      orc_neon_reg_name_quad (p->tmpreg),
-      orc_neon_reg_name (p->vars[insn->src_args[0]].alloc),
-      orc_neon_reg_name (p->vars[insn->src_args[1]].alloc));
-  x |= (p->tmpreg&0xf)<<12;
-  x |= ((p->tmpreg>>4)&0x1)<<22;
-  x |= (p->vars[insn->src_args[0]].alloc&0xf)<<16;
-  x |= ((p->vars[insn->src_args[0]].alloc>>4)&0x1)<<7;
-  x |= (p->vars[insn->src_args[1]].alloc&0xf)<<0;
-  x |= ((p->vars[insn->src_args[1]].alloc>>4)&0x1)<<5;
-  orc_arm_emit (p, x);
+  orc_uint32 x;
+  unsigned int code;
 
-  orc_neon_emit_unary (p, "vpadal.u16", 0xf3b40680,
-      p->vars[insn->dest_args[0]].alloc,
-      p->tmpreg);
+  if (p->loop_shift < 2) {
+    x = 0xf3800700;
+    ORC_ASM_CODE(p,"  vabdl.u8 %s, %s, %s\n",
+        orc_neon_reg_name_quad (p->tmpreg),
+        orc_neon_reg_name (p->vars[insn->src_args[0]].alloc),
+        orc_neon_reg_name (p->vars[insn->src_args[1]].alloc));
+    x |= (p->tmpreg&0xf)<<12;
+    x |= ((p->tmpreg>>4)&0x1)<<22;
+    x |= (p->vars[insn->src_args[0]].alloc&0xf)<<16;
+    x |= ((p->vars[insn->src_args[0]].alloc>>4)&0x1)<<7;
+    x |= (p->vars[insn->src_args[1]].alloc&0xf)<<0;
+    x |= ((p->vars[insn->src_args[1]].alloc>>4)&0x1)<<5;
+    orc_arm_emit (p, x);
+
+    ORC_ASM_CODE(p,"  vshl.i64 %s, %s, #%d\n",
+        orc_neon_reg_name (p->tmpreg),
+        orc_neon_reg_name (p->tmpreg), 64 - (16<<p->loop_shift));
+    code = NEON_BINARY(0xf2a00590, p->tmpreg, 0, p->tmpreg);
+    code |= (64 - (16<<p->loop_shift)) << 16;
+    orc_arm_emit (p, code);
+
+    orc_neon_emit_unary (p, "vpadal.u16", 0xf3b40680,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg);
+  } else {
+    x = 0xf3800700;
+    ORC_ASM_CODE(p,"  vabdl.u8 %s, %s, %s\n",
+        orc_neon_reg_name_quad (p->tmpreg),
+        orc_neon_reg_name (p->vars[insn->src_args[0]].alloc),
+        orc_neon_reg_name (p->vars[insn->src_args[1]].alloc));
+    x |= (p->tmpreg&0xf)<<12;
+    x |= ((p->tmpreg>>4)&0x1)<<22;
+    x |= (p->vars[insn->src_args[0]].alloc&0xf)<<16;
+    x |= ((p->vars[insn->src_args[0]].alloc>>4)&0x1)<<7;
+    x |= (p->vars[insn->src_args[1]].alloc&0xf)<<0;
+    x |= ((p->vars[insn->src_args[1]].alloc>>4)&0x1)<<5;
+    orc_arm_emit (p, x);
+
+    orc_neon_emit_unary (p, "vpadal.u16", 0xf3b40680,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg);
+  }
 }
 
 static void
@@ -1281,15 +1603,29 @@ orc_neon_rule_signw (OrcCompiler *p, void *user, OrcInstruction *insn)
   /* slow */
 
   orc_neon_emit_loadiw (p, p->tmpreg, 1);
-  orc_neon_emit_binary (p, "vmin.s16", 0xf2100610,
-      p->vars[insn->dest_args[0]].alloc,
-      p->tmpreg,
-      p->vars[insn->src_args[0]].alloc);
+  if (p->loop_shift < 3) {
+    orc_neon_emit_binary (p, "vmin.s16", 0xf2100610,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg,
+        p->vars[insn->src_args[0]].alloc);
+  } else {
+    orc_neon_emit_binary_quad (p, "vmin.s16", 0xf2100610,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg,
+        p->vars[insn->src_args[0]].alloc);
+  }
   orc_neon_emit_loadiw (p, p->tmpreg, -1);
-  orc_neon_emit_binary (p, "vmax.s16", 0xf2100600,
-      p->vars[insn->dest_args[0]].alloc,
-      p->tmpreg,
-      p->vars[insn->dest_args[0]].alloc);
+  if (p->loop_shift < 3) {
+    orc_neon_emit_binary (p, "vmax.s16", 0xf2100600,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg,
+        p->vars[insn->dest_args[0]].alloc);
+  } else {
+    orc_neon_emit_binary_quad (p, "vmax.s16", 0xf2100600,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg,
+        p->vars[insn->dest_args[0]].alloc);
+  }
 }
 
 static void
@@ -1308,6 +1644,20 @@ orc_neon_rule_mulhuw (OrcCompiler *p, void *user, OrcInstruction *insn)
       p->vars[insn->dest_args[0]].alloc,
       p->tmpreg, 0);
   orc_arm_emit (p, code);
+
+  if (p->loop_shift == 3) {
+    orc_neon_emit_binary_long (p, "vmull.u16",0xf3900c00,
+        p->tmpreg,
+        p->vars[insn->src_args[0]].alloc + 1,
+        p->vars[insn->src_args[1]].alloc + 1);
+    ORC_ASM_CODE(p,"  vshrn.i32 %s, %s, #%d\n",
+        orc_neon_reg_name (p->vars[insn->dest_args[0]].alloc + 1),
+        orc_neon_reg_name_quad (p->tmpreg), 16);
+    code = NEON_BINARY (0xf2900810,
+        p->vars[insn->dest_args[0]].alloc + 1,
+        p->tmpreg, 0);
+    orc_arm_emit (p, code);
+  }
 }
 
 void
@@ -1450,8 +1800,8 @@ orc_compiler_neon_register_rules (OrcTarget *target)
   orc_rule_register (rule_set, "shrsl", orc_neon_rule_shift, (void *)7);
   orc_rule_register (rule_set, "shrul", orc_neon_rule_shift, (void *)8);
 
-  orc_rule_register (rule_set, "andnb", orc_neon_rule_andn, NULL);
-  orc_rule_register (rule_set, "andnw", orc_neon_rule_andn, NULL);
-  orc_rule_register (rule_set, "andnl", orc_neon_rule_andn, NULL);
+  orc_rule_register (rule_set, "andnb", orc_neon_rule_andn, (void *)3);
+  orc_rule_register (rule_set, "andnw", orc_neon_rule_andn, (void *)2);
+  orc_rule_register (rule_set, "andnl", orc_neon_rule_andn, (void *)1);
 }
 
