@@ -478,6 +478,20 @@ sse_rule_mulubw (OrcCompiler *p, void *user, OrcInstruction *insn)
   orc_sse_emit_pmullw (p, tmp, dest);
 }
 
+static void
+sse_rule_muluwl (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  int src = p->vars[insn->src_args[1]].alloc;
+  int dest = p->vars[insn->dest_args[0]].alloc;
+  int tmp = p->tmpreg;
+
+  orc_sse_emit_punpcklwd (p, src, tmp);
+  orc_sse_emit_psrld (p, 16, tmp);
+  orc_sse_emit_punpcklwd (p, dest, dest);
+  orc_sse_emit_psrld (p, 16, dest);
+  orc_sse_emit_pmulld (p, tmp, dest);
+}
+
 static void
 sse_rule_mulswl (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
@@ -941,10 +955,54 @@ BINARY_F(addf, "addps", 0x58)
 BINARY_F(subf, "subps", 0x5c)
 BINARY_F(mulf, "mulps", 0x59)
 BINARY_F(divf, "divps", 0x5e)
-BINARY_F(maxf, "maxps", 0x5f)
-BINARY_F(minf, "minps", 0x5d)
 UNARY_F(sqrtf, "sqrtps", 0x51)
 
+static void
+sse_rule_minf (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  if (p->target_flags & ORC_TARGET_FAST_NAN) {
+    orc_sse_emit_0f (p, "minps", 0x5d,
+        p->vars[insn->src_args[1]].alloc,
+        p->vars[insn->dest_args[0]].alloc);
+  } else {
+    orc_sse_emit_movdqa (p,
+        p->vars[insn->src_args[1]].alloc,
+        p->tmpreg);
+    orc_sse_emit_0f (p, "minps", 0x5d,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg);
+    orc_sse_emit_0f (p, "minps", 0x5d,
+        p->vars[insn->src_args[1]].alloc,
+        p->vars[insn->dest_args[0]].alloc);
+    orc_sse_emit_por (p,
+        p->tmpreg,
+        p->vars[insn->dest_args[0]].alloc);
+  }
+}
+
+static void
+sse_rule_maxf (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  if (p->target_flags & ORC_TARGET_FAST_NAN) {
+    orc_sse_emit_0f (p, "maxps", 0x5f,
+        p->vars[insn->src_args[1]].alloc,
+        p->vars[insn->dest_args[0]].alloc);
+  } else {
+    orc_sse_emit_movdqa (p,
+        p->vars[insn->src_args[1]].alloc,
+        p->tmpreg);
+    orc_sse_emit_0f (p, "maxps", 0x5f,
+        p->vars[insn->dest_args[0]].alloc,
+        p->tmpreg);
+    orc_sse_emit_0f (p, "maxps", 0x5f,
+        p->vars[insn->src_args[1]].alloc,
+        p->vars[insn->dest_args[0]].alloc);
+    orc_sse_emit_por (p,
+        p->tmpreg,
+        p->vars[insn->dest_args[0]].alloc);
+  }
+}
+
 static void
 sse_rule_cmpeqf (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
@@ -1143,6 +1201,7 @@ orc_compiler_sse_register_rules (OrcTarget *target)
   REG(minul);
   REG(mulll);
   orc_rule_register (rule_set, "convsuslw", sse_rule_convsuslw, NULL);
+  orc_rule_register (rule_set, "muluwl", sse_rule_muluwl, NULL);
 
   /* SSE 4.2 -- no rules */
 
