@@ -470,8 +470,27 @@ orc_x86_emit_add_reg_memoffset (OrcCompiler *compiler, int size, int reg1,
 }
 
 void
-orc_x86_emit_add_imm_reg (OrcCompiler *compiler, int size, int value, int reg)
+orc_x86_emit_add_imm_reg (OrcCompiler *compiler, int size, int value, int reg, orc_bool record)
 {
+  if (!record) {
+    if (size == 4 && !compiler->is_64bit) {
+      ORC_ASM_CODE(compiler,"  lea %d(%%%s), %%%s\n", value,
+          orc_x86_get_regname(reg), orc_x86_get_regname(reg));
+      orc_x86_emit_rex(compiler, size, 0, 0, reg);
+      *compiler->codeptr++ = 0x8d;
+      orc_x86_emit_modrm_memoffset (compiler, reg, value, reg);
+      return;
+    }
+    if (size == 8 && compiler->is_64bit) {
+      ORC_ASM_CODE(compiler,"  lea %d(%%%s), %%%s\n", value,
+          orc_x86_get_regname_64(reg), orc_x86_get_regname_64(reg));
+      orc_x86_emit_rex(compiler, size, reg, 0, reg);
+      *compiler->codeptr++ = 0x8d;
+      orc_x86_emit_modrm_memoffset (compiler, reg, value, reg);
+      return;
+    }
+  }
+
   if (size == 2) {
     ORC_ASM_CODE(compiler,"  addw $%d, %%%s\n", value, orc_x86_get_regname_16(reg));
     *compiler->codeptr++ = 0x66;
@@ -831,8 +850,8 @@ orc_x86_do_fixups (OrcCompiler *compiler)
       unsigned char *ptr = compiler->fixups[i].ptr;
       int diff;
 
-      diff = ((int8_t)ptr[0]) + (label - ptr);
-      if (diff != (int8_t)diff) {
+      diff = ((orc_int8)ptr[0]) + (label - ptr);
+      if (diff != (orc_int8)diff) {
         ORC_COMPILER_ERROR(compiler, "short jump too long");
       }
 
@@ -878,6 +897,7 @@ orc_x86_emit_prologue (OrcCompiler *compiler)
       orc_x86_emit_push (compiler, 4, X86_EBX);
     }
   }
+
 }
 
 void
@@ -888,7 +908,7 @@ orc_x86_emit_epilogue (OrcCompiler *compiler)
     for(i=15;i>=0;i--){
       if (compiler->used_regs[ORC_GP_REG_BASE+i] &&
           compiler->save_regs[ORC_GP_REG_BASE+i]) {
-        orc_x86_emit_push (compiler, 8, ORC_GP_REG_BASE+i);
+        orc_x86_emit_pop (compiler, 8, ORC_GP_REG_BASE+i);
       }
     }
   } else {
