@@ -17,15 +17,23 @@ void output_code_test (OrcProgram *p, FILE *output);
 void output_code_backup (OrcProgram *p, FILE *output);
 void output_code_no_orc (OrcProgram *p, FILE *output);
 void output_code_assembly (OrcProgram *p, FILE *output);
-static void print_exec_header (FILE *output);
 static char * get_barrier (const char *s);
 static const char * my_basename (const char *s);
 
 int verbose = 0;
 int error = 0;
+int compat;
 
 char *target = "sse";
 
+#define ORC_VERSION(a,b,c,d) ((a)*1000000 + (b)*10000 + (c)*100 + (d))
+#define REQUIRE(a,b,c,d) do { \
+  if (ORC_VERSION((a),(b),(c),(d)) > compat) { \
+    fprintf(stderr, "Feature used that is incompatible with --compat\n"); \
+    exit (1); \
+  } \
+} while (0)
+
 enum {
   MODE_IMPL,
   MODE_HEADER,
@@ -40,17 +48,18 @@ void help (void)
   printf("  orcc [OPTION...] INPUT_FILE\n");
   printf("\n");
   printf("Help Options:\n");
-  printf("  -?, --help                      Show help options\n");
+  printf("  -h, --help              Show help options\n");
   printf("\n");
   printf("Application Options:\n");
-  printf("  -v, --verbose                   Output more information\n");
-  printf("  -o, --output FILE               Write output to FILE\n");
-  printf("  --implementation                Produce C code implementing functions\n");
-  printf("  --header                        Produce C header for functions\n");
-  printf("  --test                          Produce test code for functions\n");
-  printf("  --assembly                      Produce assembly code for functions\n");
-  printf("  --include FILE                  Generate #include <FILE> in code\n");
-  printf("  --target TARGET                 Generate assembly for TARGET\n");
+  printf("  -v, --verbose           Output more information\n");
+  printf("  -o, --output FILE       Write output to FILE\n");
+  printf("  --implementation        Produce C code implementing functions\n");
+  printf("  --header                Produce C header for functions\n");
+  printf("  --test                  Produce test code for functions\n");
+  printf("  --assembly              Produce assembly code for functions\n");
+  printf("  --include FILE          Add #include <FILE> to code\n");
+  printf("  --target TARGET         Generate assembly for TARGET\n");
+  printf("  --compat VERSION        Generate code compatible with Orc version VERSION\n");
   printf("\n");
 
   exit (0);
@@ -66,6 +75,7 @@ main (int argc, char *argv[])
   char *output_file = NULL;
   char *input_file = NULL;
   char *include_file = NULL;
+  char *compat_version = VERSION;
   FILE *output;
 
   orc_init ();
@@ -111,6 +121,13 @@ main (int argc, char *argv[])
     } else if (strcmp(argv[i], "--version") == 0) {
       printf("Orc Compiler " PACKAGE_VERSION "\n");
       exit (0);
+    } else if (strcmp(argv[i], "--compat") == 0) {
+      if (i+1 < argc) {
+        compat_version = argv[i+1];
+        i++;
+      } else {
+        help();
+      }
     } else if (strncmp(argv[i], "-", 1) == 0) {
       printf("Unknown option: %s\n", argv[i]);
       exit (1);
@@ -134,6 +151,25 @@ main (int argc, char *argv[])
     exit (1);
   }
 
+  if (compat_version) {
+    int major, minor, micro, nano = 0;
+    int n;
+
+    n = sscanf (compat_version, "%d.%d.%d.%d", &major, &minor, &micro, &nano);
+
+    if (n < 3) {
+      printf("Unknown version \"%s\"\n", compat_version);
+      exit (1);
+    }
+
+    compat = ORC_VERSION(major,minor,micro,nano);
+    if (compat < ORC_VERSION(0,4,5,0)) {
+      printf("Compatibility version \"%s\" not supported.  Minimum 0.4.5\n",
+          compat_version);
+      exit (1);
+    }
+  }
+
   if (output_file == NULL) {
     if (mode == MODE_IMPL) {
       output_file = "out.c";
@@ -168,14 +204,13 @@ main (int argc, char *argv[])
     fprintf(output, "#endif\n");
     fprintf(output, "#ifndef DISABLE_ORC\n");
     fprintf(output, "#include <orc/orc.h>\n");
-    fprintf(output, "#else\n");
-    fprintf(output, "#include <stdint.h>\n");
-    print_exec_header (output);
     fprintf(output, "#endif\n");
     if (include_file) {
       fprintf(output, "#include <%s>\n", include_file);
     }
     fprintf(output, "\n");
+    fprintf(output, "%s", orc_target_c_get_typedefs ());
+    fprintf(output, "\n");
     for(i=0;i<n;i++){
       output_code_header (programs[i], output);
     }
@@ -200,6 +235,8 @@ main (int argc, char *argv[])
     fprintf(output, "extern \"C\" {\n");
     fprintf(output, "#endif\n");
     fprintf(output, "\n");
+    fprintf(output, "%s", orc_target_c_get_typedefs ());
+    fprintf(output, "\n");
     for(i=0;i<n;i++){
       output_code_header (programs[i], output);
     }
@@ -220,6 +257,7 @@ main (int argc, char *argv[])
       fprintf(output, "#include <%s>\n", include_file);
     }
     fprintf(output, "\n");
+    fprintf(output, "%s", orc_target_c_get_typedefs ());
     fprintf(output, "%s", orc_target_get_asm_preamble ("c"));
     fprintf(output, "\n");
     for(i=0;i<n;i++){
@@ -227,13 +265,45 @@ main (int argc, char *argv[])
       output_code_backup (programs[i], output);
     }
     fprintf(output, "\n");
+    fprintf(output, "static int quiet = 0;\n");
+    fprintf(output, "static int benchmark = 0;\n");
+    fprintf(output, "\n");
+    fprintf(output, "static void help (const char *argv0)\n");
+    fprintf(output, "{\n");
+    fprintf(output, "  printf(\"Usage:\\n\");\n");
+    fprintf(output, "  printf(\"  %%s [OPTION]\\n\", argv0);\n");
+    fprintf(output, "  printf(\"Help Options:\\n\");\n");
+    fprintf(output, "  printf(\"  -h, --help          Show help options\\n\");\n");
+    fprintf(output, "  printf(\"Application Options:\\n\");\n");
+    fprintf(output, "  printf(\"  -b, --benchmark     Run benchmark and show results\\n\");\n");
+    fprintf(output, "  printf(\"  -q, --quiet         Don't output anything except on failures\\n\");\n");
+    fprintf(output, "\n");
+    fprintf(output, "  exit(0);\n");
+    fprintf(output, "}\n");
+    fprintf(output, "\n");
     fprintf(output, "int\n");
     fprintf(output, "main (int argc, char *argv[])\n");
     fprintf(output, "{\n");
     fprintf(output, "  int error = FALSE;\n");
+    fprintf(output, "  int i;\n");
     fprintf(output, "\n");
     fprintf(output, "  orc_test_init ();\n");
     fprintf(output, "\n");
+    fprintf(output, "  for(i=1;i<argc;i++) {\n");
+    fprintf(output, "    if (strcmp(argv[i], \"--help\") == 0 ||\n");
+    fprintf(output, "      strcmp(argv[i], \"-h\") == 0) {\n");
+    fprintf(output, "      help(argv[0]);\n");
+    fprintf(output, "    } else if (strcmp(argv[i], \"--quiet\") == 0 ||\n");
+    fprintf(output, "      strcmp(argv[i], \"-q\") == 0) {\n");
+    fprintf(output, "      quiet = 1;\n");
+    fprintf(output, "      benchmark = 0;\n");
+    fprintf(output, "    } else if (strcmp(argv[i], \"--benchmark\") == 0 ||\n");
+    fprintf(output, "      strcmp(argv[i], \"-b\") == 0) {\n");
+    fprintf(output, "      benchmark = 1;\n");
+    fprintf(output, "      quiet = 0;\n");
+    fprintf(output, "    }\n");
+    fprintf(output, "  }\n");
+    fprintf(output, "\n");
     for(i=0;i<n;i++){
       output_code_test (programs[i], output);
     }
@@ -258,77 +328,6 @@ main (int argc, char *argv[])
 }
 
 
-static void
-print_exec_header (FILE *output)
-{
-#if 0
-  fprintf(output,
-      "typedef struct _OrcExecutor OrcExecutor;\n"
-      "typedef struct _OrcProgram OrcProgram;\n"
-      "#define ORC_N_VARIABLES 64\n"
-      "struct _OrcExecutor {\n"
-      "  OrcProgram *program;\n"
-      "  int n;\n"
-      "  int counter1;\n"
-      "  int counter2;\n"
-      "  int counter3;\n"
-      "  void *arrays[ORC_N_VARIABLES];\n"
-      "  int params[ORC_N_VARIABLES];\n"
-      "  int accumulators[4];\n"
-      "};\n"
-      "\n"
-      "enum {\n"
-      "  ORC_VAR_D1,\n"
-      "  ORC_VAR_D2,\n"
-      "  ORC_VAR_D3,\n"
-      "  ORC_VAR_D4,\n"
-      "  ORC_VAR_S1,\n"
-      "  ORC_VAR_S2,\n"
-      "  ORC_VAR_S3,\n"
-      "  ORC_VAR_S4,\n"
-      "  ORC_VAR_S5,\n"
-      "  ORC_VAR_S6,\n"
-      "  ORC_VAR_S7,\n"
-      "  ORC_VAR_S8,\n"
-      "  ORC_VAR_A1,\n"
-      "  ORC_VAR_A2,\n"
-      "  ORC_VAR_A3,\n"
-      "  ORC_VAR_A4,\n"
-      "  ORC_VAR_C1,\n"
-      "  ORC_VAR_C2,\n"
-      "  ORC_VAR_C3,\n"
-      "  ORC_VAR_C4,\n"
-      "  ORC_VAR_C5,\n"
-      "  ORC_VAR_C6,\n"
-      "  ORC_VAR_C7,\n"
-      "  ORC_VAR_C8,\n"
-      "  ORC_VAR_P1,\n"
-      "  ORC_VAR_P2,\n"
-      "  ORC_VAR_P3,\n"
-      "  ORC_VAR_P4,\n"
-      "  ORC_VAR_P5,\n"
-      "  ORC_VAR_P6,\n"
-      "  ORC_VAR_P7,\n"
-      "  ORC_VAR_P8,\n"
-      "  ORC_VAR_T1,\n"
-      "  ORC_VAR_T2,\n"
-      "  ORC_VAR_T3,\n"
-      "  ORC_VAR_T4,\n"
-      "  ORC_VAR_T5,\n"
-      "  ORC_VAR_T6,\n"
-      "  ORC_VAR_T7,\n"
-      "  ORC_VAR_T8,\n"
-      "  ORC_VAR_T9,\n"
-      "  ORC_VAR_T10,\n"
-      "  ORC_VAR_T11,\n"
-      "  ORC_VAR_T12,\n"
-      "  ORC_VAR_T13,\n"
-      "  ORC_VAR_T14,\n"
-      "  ORC_VAR_T15\n"
-      "};\n");
-#endif
-}
-
 static char *
 get_barrier (const char *s)
 {
@@ -350,39 +349,6 @@ get_barrier (const char *s)
   return barrier;
 }
 
-#if 0
-static void
-print_defines (FILE *output)
-{
-  fprintf(output,
-    "#define ORC_CLAMP(x,a,b) ((x)<(a) ? (a) : ((x)>(b) ? (b) : (x)))\n"
-    "#define ORC_ABS(a) ((a)<0 ? -(a) : (a))\n"
-    "#define ORC_MIN(a,b) ((a)<(b) ? (a) : (b))\n"
-    "#define ORC_MAX(a,b) ((a)>(b) ? (a) : (b))\n"
-    "#define ORC_SB_MAX 127\n"
-    "#define ORC_SB_MIN (-1-ORC_SB_MAX)\n"
-    "#define ORC_UB_MAX 255\n"
-    "#define ORC_UB_MIN 0\n"
-    "#define ORC_SW_MAX 32767\n"
-    "#define ORC_SW_MIN (-1-ORC_SW_MAX)\n"
-    "#define ORC_UW_MAX 65535\n"
-    "#define ORC_UW_MIN 0\n"
-    "#define ORC_SL_MAX 2147483647\n"
-    "#define ORC_SL_MIN (-1-ORC_SL_MAX)\n"
-    "#define ORC_UL_MAX 4294967295U\n"
-    "#define ORC_UL_MIN 0\n"
-    "#define ORC_CLAMP_SB(x) ORC_CLAMP(x,ORC_SB_MIN,ORC_SB_MAX)\n"
-    "#define ORC_CLAMP_UB(x) ORC_CLAMP(x,ORC_UB_MIN,ORC_UB_MAX)\n"
-    "#define ORC_CLAMP_SW(x) ORC_CLAMP(x,ORC_SW_MIN,ORC_SW_MAX)\n"
-    "#define ORC_CLAMP_UW(x) ORC_CLAMP(x,ORC_UW_MIN,ORC_UW_MAX)\n"
-    "#define ORC_CLAMP_SL(x) ORC_CLAMP(x,ORC_SL_MIN,ORC_SL_MAX)\n"
-    "#define ORC_CLAMP_UL(x) ORC_CLAMP(x,ORC_UL_MIN,ORC_UL_MAX)\n"
-    "#define ORC_SWAP_W(x) ((((x)&0xff)<<8) | (((x)&0xff00)>>8))\n"
-    "#define ORC_SWAP_L(x) ((((x)&0xff)<<24) | (((x)&0xff00)<<8) | (((x)&0xff0000)>>8) | (((x)&0xff000000)>>24))\n"
-    "#define ORC_PTR_OFFSET(ptr,offset) ((void *)(((unsigned char *)(ptr)) + (offset)))\n");
-}
-#endif
-
 static char *
 read_file (const char *filename)
 {
@@ -467,7 +433,7 @@ output_prototype (OrcProgram *p, FILE *output)
         fprintf(output, "%s * %s", var->type_name,
             varnames[ORC_VAR_D1 + i]);
       } else {
-        fprintf(output, "uint%d_t * %s", var->size*8,
+        fprintf(output, "orc_uint%d * %s", var->size*8,
             varnames[ORC_VAR_D1 + i]);
       }
       if (p->is_2d) {
@@ -484,7 +450,7 @@ output_prototype (OrcProgram *p, FILE *output)
         fprintf(output, "%s * %s", var->type_name,
             varnames[ORC_VAR_A1 + i]);
       } else {
-        fprintf(output, "uint%d_t * %s", var->size*8,
+        fprintf(output, "orc_uint%d * %s", var->size*8,
             varnames[ORC_VAR_A1 + i]);
       }
       need_comma = TRUE;
@@ -498,7 +464,7 @@ output_prototype (OrcProgram *p, FILE *output)
         fprintf(output, "const %s * %s", var->type_name,
             varnames[ORC_VAR_S1 + i]);
       } else {
-        fprintf(output, "const uint%d_t * %s", var->size*8,
+        fprintf(output, "const orc_uint%d * %s", var->size*8,
             varnames[ORC_VAR_S1 + i]);
       }
       if (p->is_2d) {
@@ -511,7 +477,12 @@ output_prototype (OrcProgram *p, FILE *output)
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
       if (need_comma) fprintf(output, ", ");
-      fprintf(output, "int %s", varnames[ORC_VAR_P1 + i]);
+      if (var->is_float_param) {
+        REQUIRE(0,4,5,1);
+      }
+      fprintf(output, "%s %s",
+          var->is_float_param ? "float" : "int",
+          varnames[ORC_VAR_P1 + i]);
       need_comma = TRUE;
     }
   }
@@ -660,7 +631,11 @@ output_code (OrcProgram *p, FILE *output)
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
-      fprintf(output, "      orc_program_add_parameter (p, %d, \"%s\");\n",
+      if (var->is_float_param) {
+        REQUIRE(0,4,5,1);
+      }
+      fprintf(output, "      orc_program_add_parameter%s (p, %d, \"%s\");\n",
+          var->is_float_param ? "_float" : "",
           var->size, varnames[ORC_VAR_P1 + i]);
     }
   }
@@ -731,8 +706,18 @@ output_code (OrcProgram *p, FILE *output)
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
-      fprintf(output, "  ex->params[%s] = %s;\n",
-          enumnames[ORC_VAR_P1 + i], varnames[ORC_VAR_P1 + i]);
+      if (var->is_float_param) {
+        REQUIRE(0,4,5,1);
+        fprintf(output, "  {\n");
+        fprintf(output, "    orc_union32 tmp;\n");
+        fprintf(output, "    tmp.f = %s;\n", varnames[ORC_VAR_P1 + i]);
+        fprintf(output, "    ex->params[%s] = tmp.i;\n",
+            enumnames[ORC_VAR_P1 + i]);
+        fprintf(output, "  }\n");
+      } else {
+        fprintf(output, "  ex->params[%s] = %s;\n",
+            enumnames[ORC_VAR_P1 + i], varnames[ORC_VAR_P1 + i]);
+      }
     }
   }
   fprintf(output, "\n");
@@ -762,7 +747,8 @@ output_code_test (OrcProgram *p, FILE *output)
   fprintf(output, "    OrcProgram *p = NULL;\n");
   fprintf(output, "    int ret;\n");
   fprintf(output, "\n");
-  fprintf(output, "    printf (\"%s:\\n\");\n", p->name);
+  fprintf(output, "    if (!quiet)");
+  fprintf(output, "      printf (\"%s:\\n\");\n", p->name);
   fprintf(output, "    p = orc_program_new ();\n");
   if (p->constant_n != 0) {
     fprintf(output, "      orc_program_set_constant_n (p, %d);\n",
@@ -814,7 +800,11 @@ output_code_test (OrcProgram *p, FILE *output)
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
-      fprintf(output, "    orc_program_add_parameter (p, %d, \"%s\");\n",
+      if (var->is_float_param) {
+        REQUIRE(0,4,5,1);
+      }
+      fprintf(output, "    orc_program_add_parameter%s (p, %d, \"%s\");\n",
+          var->is_float_param ? "_float" : "",
           var->size, varnames[ORC_VAR_P1 + i]);
     }
   }
@@ -840,22 +830,32 @@ output_code_test (OrcProgram *p, FILE *output)
     }
   }
 
-  fprintf(output, "\n");
-  fprintf(output, "    printf (\"  cycles (backup):   \");\n");
-  fprintf(output, "    orc_test_performance (p, ORC_TEST_FLAGS_BACKUP);\n");
   fprintf(output, "\n");
   fprintf(output, "    ret = orc_test_compare_output_backup (p);\n");
   fprintf(output, "    if (!ret) {\n");
   fprintf(output, "      error = TRUE;\n");
+  fprintf(output, "    } else if (!quiet) {\n");
+  fprintf(output, "      printf (\"    backup function  :   PASSED\\n\");\n");
+  fprintf(output, "    }\n");
+  fprintf(output, "\n");
+  fprintf(output, "    if (benchmark) {\n");
+  fprintf(output, "      printf (\"    cycles (backup)  :   \");\n");
+  fprintf(output, "      orc_test_performance (p, ORC_TEST_FLAGS_BACKUP);\n");
   fprintf(output, "    }\n");
   fprintf(output, "\n");
   fprintf(output, "    ret = orc_test_compare_output (p);\n");
-  fprintf(output, "    if (!ret) {\n");
+  fprintf(output, "    if (ret == ORC_TEST_INDETERMINATE && !quiet) {\n");
+  fprintf(output, "      printf (\"    compiled function:   COMPILE FAILED\\n\");\n");
+  fprintf(output, "    } else if (!ret) {\n");
   fprintf(output, "      error = TRUE;\n");
+  fprintf(output, "    } else if (!quiet) {\n");
+  fprintf(output, "      printf (\"    compiled function:   PASSED\\n\");\n");
   fprintf(output, "    }\n");
   fprintf(output, "\n");
-  fprintf(output, "    printf (\"  cycles (compiled): \");\n");
-  fprintf(output, "    orc_test_performance (p, 0);\n");
+  fprintf(output, "    if (benchmark) {\n");
+  fprintf(output, "      printf (\"    cycles (compiled):   \");\n");
+  fprintf(output, "      orc_test_performance (p, 0);\n");
+  fprintf(output, "    }\n");
   fprintf(output, "\n");
   fprintf(output, "    orc_program_free (p);\n");
   fprintf(output, "  }\n");
